# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualisePoint3 shows a basic point cloud containing every point in the dict
with a timestamp within the time window. 
The world (everywhere the points reach in the whole data) is compressed into
a rectangle which is then viewed from the outside as if the world reference
frame were aligned with a standard camera reference frame,
so z increases away from the viewer, x increases to the right,
and y in a downwards direction.
The user has a choice to apply perspective (scale decreases away from the 
viewer) or not.
There are also pitch and yaw controls - this alters the basic viewpoint of
the viewer described above.
"""

import numpy as np

# Local imports
from .visualiserBase import Visualiser


class VisualiserPoint3(Visualiser):
    renderX = 300  # TODO Hardcoded
    renderY = 300
    labels = None
    data_type = 'point3'

    def __init__(self, data):
        self.set_data(data)
        self.smallestRenderDim = min(self.renderX, self.renderY)

    '''
    Offset and scale the pose translations so that they all fit into the volume:
        x [-0.5:0.5]
        y[-0.5:0.5]
        z[1:2]
    '''

    def set_data(self, data):
        # scale and offset point data so that it remains proportional 
        # but stays in the range 0-1 for all dimensions
        pointX = data['point'][:, 0]
        pointY = data['point'][:, 1]
        pointZ = data['point'][:, 2]
        minX = np.min(pointX)
        maxX = np.max(pointX)
        minY = np.min(pointY)
        maxY = np.max(pointY)
        minZ = np.min(pointZ)
        maxZ = np.max(pointZ)
        centreX = (minX + maxX) / 2
        centreY = (minY + maxY) / 2
        centreZ = (minZ + maxZ) / 2
        largestDim = max(maxX - minX, maxY - minY, maxZ - minZ)
        if largestDim == 0:
            largestDim = 1

        pointScaled = np.empty_like(data['point'])
        pointScaled[:, 0] = pointX - centreX
        pointScaled[:, 1] = pointY - centreY
        pointScaled[:, 2] = pointZ - centreZ
        pointScaled = pointScaled / largestDim + 0.5
        internalData = {'ts': data['ts'],
                        'point': pointScaled
                        }
        self.__data = internalData

    def project3dTo2d(self, x=0, y=0, z=0, **kwargs):
        smallestRenderDim = kwargs.get('smallestRenderDim', 1)
        windowFill = kwargs.get('windowFill', 0.9)
        if kwargs.get('perspective', True):
            # Move z out by 1, so that the data is between 1 and 2 distant in z
            # x and y are in range 0-1, so they get shifted to be centred around 0 during this operation
            x = (x - 0.5) / (z + 1) + 0.5
            y = (y - 0.5) / (z + 1) + 0.5
        x = (x * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        y = (y * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        x = x.astype(int)
        y = y.astype(int)
        return x, y

    def point_to_image(self, point, image, **kwargs):
        if point is None:
            return image
        # Unpack
        pointX = point[0]
        pointY = point[1]
        pointZ = point[2]
        # Project the location
        projX, projY = self.project3dTo2d(x=pointX, y=pointY, z=pointZ,
                                          smallestRenderDim=self.smallestRenderDim, **kwargs)
        try:
            image[projY, projX, :] = 255
        except IndexError:  # perspective or other projection issues cause out of bounds? ignore
            pass
        return image

    def get_frame(self, time, timeWindow, **kwargs):
        data = self.__data
        if data is None:
            print('Warning: data is not set')
            return np.zeros((1, 1), dtype=np.uint8)  # This should not happen
        image = np.zeros((self.renderY, self.renderX, 3), dtype=np.uint8)
        # Put a grey box around the edge of the image
        image[0, :, :] = 128
        image[-1, :, :] = 128
        image[:, 0, :] = 128
        image[:, -1, :] = 128
        # Put a grey crosshair in the centre of the image
        rY = self.renderY
        rX = self.renderY
        chp = 20  # Cross Hair Proportion for following expression
        image[int(rY / 2 - rY / chp): int(rY / 2 + rY / chp), int(rX / 2), :] = 128
        image[int(rY / 2), int(rX / 2 - rX / chp): int(rX / 2 + rX / chp), :] = 128
        firstIdx = np.searchsorted(data['ts'], time - timeWindow)
        lastIdx = np.searchsorted(data['ts'], time + timeWindow)
        points = data['point'][firstIdx:lastIdx, :]
        # Use yaw and pitch sliders to transform points
        yaw = -kwargs.get('yaw', 0) / 180 * np.pi
        pitch = kwargs.get('pitch', 0) / 180 * np.pi
        roll = 0
        cosA = np.cos(roll)
        cosB = np.cos(yaw)
        cosC = np.cos(pitch)
        sinA = np.sin(roll)
        sinB = np.sin(yaw)
        sinC = np.sin(pitch)
        rotMat = np.array([[cosA * cosB, cosA * sinB * sinC - sinA * cosC, cosA * sinB * cosC + sinA * sinC],
                           [sinA * cosB, sinA * sinB * sinC + cosA * cosC, sinA * sinB * cosC - cosA * sinC],
                           [-sinB, cosB * sinC, cosB * cosC]],
                          dtype=float)
        points = points - 0.5
        points = np.matmul(rotMat, points.transpose()).transpose()
        points = points + 0.5

        for row in points:
            image = self.point_to_image(row, image, **kwargs)

            # Allow for arbitrary post-production on image with a callback
        # TODO: as this is boilerplate, it could be pushed into pie syntax ...
        if kwargs.get('callback', None) is not None:
            kwargs['image'] = image
            image = kwargs['callback'](**kwargs)
        return image

    def get_dims(self):
        return self.renderX, self.renderY

    def get_colorfmt(self):
        return 'rgb'

    def get_settings(self):
        settings = {'perspective': {'type': 'boolean',
                                    'default': True
                                    },
                    'yaw': {'type': 'range',
                            'default': 0,
                            'min': -90,
                            'max': 90,
                            'step': 1
                            },
                    'pitch': {'type': 'range',
                              'default': 0,
                              'min': -90,
                              'max': 90,
                              'step': 1
                              }}

        return settings



# -*- coding: utf-8 -*-

'''
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotFrame takes 'inDict' - a dictionary containing imported frame data 
(or a higher level container, in which attempts to descend and call itself) 
as created by importAe, and creates a series of images from selected
frames.
The number of subplots is given by the numPlots parameter.
'distributeBy' can either be 'time' or 'events', to decide how the points 
around which data is rendered are chosen. 
The frame events are then chosen as those nearest to the time points.
If the 'distributeBy' is 'time' then if the further parameters 'minTime' 
and 'maxTime' are used then the time window used is only between
those limits.
Params include:
numPlots, distributeBy, minTime, maxTime, flipVertical, flipHorizontal, transpose
'''

import numpy as np
import matplotlib.pyplot as plt
from math import log10, floor

def roundToSf(x, sig=3):
    try:
        return round(x, sig-int(floor(log10(abs(x))))-1)
    except ValueError: # log of zero
        return 0

def plotFrame(inDicts, **kwargs):
    if isinstance(inDicts, list):
        for inDict in inDicts:
            plotFrame(inDict, **kwargs)
        return
    else:
        inDict = inDicts
    if not isinstance(inDict, dict):
        return
    if 'ts' not in inDict:
        title = kwargs.pop('title', '')
        if 'info' in inDict and isinstance(inDict, dict):
            fileName = inDict['info'].get('filePathOrName')
            if fileName is not None:
                print('plotFrame was called for file ' + fileName)
                title = (title + ' ' + fileName).lstrip()
        for key in inDict.keys():
            kwargs['title'] = (title + ' ' + key).lstrip()
            plotFrame(inDict[key], **kwargs)
        return
    # From this point onwards, it's a data-type container
    if 'frames' not in inDict:
        return
    # From this point onwards, it's a frame data-type container
    distributeBy = kwargs.get('distributeBy', 'time').lower()
    numPlots = kwargs.get('numPlots', 6)
    
    ts = inDict['ts']
    frames = inDict['frames']
    numFrames = len(ts)
    if numFrames < numPlots:
        numPlots = numFrames

    if numFrames == numPlots:
        distributeBy = 'events'
    
    # Distribute plots in a raster with a 3:4 ratio
    numPlotsX = int(np.round(np.sqrt(numPlots / 3 * 4)))
    numPlotsY = int(np.ceil(numPlots / numPlotsX))
    
    minTime = kwargs.get('startTime', kwargs.get('minTime', kwargs.get('beginTime', ts[0])))
    maxTime = kwargs.get('stopTime', kwargs.get('maxTime', kwargs.get('endTime', ts[-1])))

    if distributeBy == 'time':
        totalTime = maxTime - minTime
        timeStep = totalTime / numPlots
        timePoints = np.arange(minTime + timeStep * 0.5, maxTime, timeStep)
    else: # distribute by event number
        framesPerStep = numFrames / numPlots
        timePoints = ts(int(np.ceil(np.arange(framesPerStep * 0.5, numFrames, framesPerStep))))

    fig, axes = plt.subplots(numPlotsY, numPlotsX)
    fig.suptitle(kwargs.get('title', ''))
    
    axes = axes.flatten().tolist()
    for ax, timePoint in zip(axes, timePoints):

        # Find eventIndex nearest to timePoint
        frameIdx = np.searchsorted(ts, timePoint)
        frame = frames[frameIdx]
        if kwargs.get('transpose', False):
            frame = np.transpose(frame)
        if kwargs.get('flipVertical', False):
            frame = np.flip(frame, axis=0)
        if kwargs.get('flipHorizontal', False):
            frame = np.flip(frame, axis=1)
        ax.imshow(frame, cmap='gray')
        ax.grid(b=False)
        ax.set_title('Time: ' + str(roundToSf(timePoint)) + ' s; frame number: ' + str(frameIdx))

#%%
'''
    Optional extra - not including it by default because I don't want this 
    extra dependency, but this is a quick way to see frame data as a video
'''

'''       
from imageio import mimwrite
def framesToGif(framesDict, **kwargs):
    ts = framesDict['ts']
    frames = framesDict['frames']
    outputFilePathAndName = kwargs.get('outputFilePathAndName', 'framesAsMovie.mp4')
    frameRate = len(ts) / (ts[-1] - ts[0])
    framesExpanded = [np.expand_dims(f, 0) for f in frames]
    framesAllArray = np.concatenate(framesExpanded, 0)
    mimwrite(outputFilePathAndName, framesAllArray , fps = int(frameRate))
'''




# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Contains general functions for text summaries of the contents of dicts which result 
from imports using importAe function
"""

from math import log10, floor
import numpy as np
import pprint
pp = pprint.PrettyPrinter(indent=12)

# Round to 3 s.f.
def sf3(x):
    if x and isinstance(x, (int, float, complex)) and not isinstance(x, bool):
        return round(x, -int(floor(log10(abs(x)))) + 2)
    else:
        return x

def fieldMinMax(dataTypeDict, fieldName):
    if fieldName in dataTypeDict:
        field = dataTypeDict[fieldName]
        if type(field) == np.ndarray:
            if field.shape[0] > 1:
                # Handle 2D arrays, e.g. an array containing x, y, z in columns
                try:
                    if len(field.shape) > 1:
                        for dim1Idx in range(field.shape[1]):
                            print('            ', 
                                  sf3(np.min(field[:, dim1Idx])), 
                                  ' >= ', fieldName, ' - col ', dim1Idx,  
                                  ' >= ', sf3(np.max(field[:, dim1Idx])))
                    else:
                        print('            ', 
                              sf3(np.min(field)), 
                              ' >= ', fieldName,
                              ' >= ', sf3(np.max(field)))
                except ValueError:
                    print('            ', fieldName, ' contains data error!')  

def info(containers, **kwargs):
    if not isinstance(containers, list):
        containers = [containers]
    for container in containers:
        pp.pprint(container['info'])
        for channelName in container['data']:
            print('    Channel: ' + channelName)
            for dataType in container['data'][channelName]:
                print('        DataType: ' + dataType)
                dataTypeDict = container['data'][channelName][dataType]
                if 'ts' in dataTypeDict:
                    print('            Num events: ', len(dataTypeDict['ts']))
                    fieldMinMax(dataTypeDict, 'ts')
                    if 'tsOffset' in dataTypeDict:
                        print('            Ts offset: ', dataTypeDict['tsOffset'])
                    for fieldName in dataTypeDict.keys():
                        if fieldName not in ['ts', 'tsOffset']:
                            fieldMinMax(dataTypeDict, fieldName)
                else:
                    pp.pprint(dataTypeDict)
                print()
            print()
def infoTs(containers, **kwargs):
    if not isinstance(containers, list):
        containers = [containers]
    for container in containers:
        print(container['info'])
        for channelName in container['data']:
            print('    Channel: ' + channelName)
            for dataType in container['data'][channelName]:
                print('        DataType: ' + dataType)
                dataTypeDict = container['data'][channelName][dataType]
                if 'ts' in dataTypeDict:
                    fieldMinMax(dataTypeDict, 'ts')
                if 'tsOffset' in dataTypeDict:
                    print('            Ts offset: ', dataTypeDict['tsOffset'])

#%% Legacy function names

def dict_keys_print(d, indent):
    print(' ' * (4 * indent - 2) + '{ ', end='')
    first = True
    for key, value in d.items():
        if first:
            print(str(key), end='')
            first = False
        else:
            print(' ' * 4 * indent + str(key), end='')
        if isinstance(value, dict):
            print(':')
            dict_keys_print(value, indent + 1)
        else:
            print(',')
            continue
    print(' ' * (4 * indent - 2) + '}')

def infoForImportedDicts(container, **kwargs):
    info(container, **kwargs)

def infoTsForImportedDicts(container, **kwargs):
    infoTs(container, **kwargs)


# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiserOpticFlow - for dense optic flow maps ...
"""

import numpy as np
import math

# Local imports
from .visualiserBase import Visualiser

# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def findNearest(array, value):
    idx = np.searchsorted(array, value) # side="left" param is the default
    if idx > 0 and ( \
            idx == len(array) or \
            math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx


class VisualiserOpticFlow(Visualiser):

    data_type = 'flowMap'

    def __init__(self, data):
        self.colorwheel = self.make_colorwheel()  # shape [55x3]
        self.set_data(data)

    def set_data(self, data):
        self.__data = {}
        self.__data.update(data)

    def get_colorfmt(self):
        return 'rgb'

    def get_default_image(self):
        x, y = self.get_dims()
        # Return an x,y,3 by default i.e. rgb, for safety, since in the absence of data we may not know how the texture's colorfmt is set
        return np.ones((x, y, 3), dtype=np.uint8) * 128 # TODO: Hardcoded midway (grey) value

    # TODO: There can be methods which better choose the best frame, or which create a visualisation which
    # respects the time_window parameter
    def get_frame(self, time, timeWindow, **kwargs):
        data = self.__data
        if time < data['ts'][0] - timeWindow / 2 or time > data['ts'][-1] + timeWindow / 2:
            # Gone off the end of the frame data
            image = self.get_default_image()
        else:
            frameIdx = findNearest(data['ts'], time)
            image = self.flow_to_color(data['flowMaps'][frameIdx])

        return image

    def get_dims(self):
        try:
            data = self.__data
        except AttributeError: # data hasn't been set yet
            return 1, 1
        x = data['dimX'] if 'dimX' in data else data['flowMaps'][0].shape[1]
        y = data['dimY'] if 'dimY' in data else data['flowMaps'][0].shape[0]
        return x, y

    ''' The methods make_colorwheel, flow_to_color and flow_uv_to_colors are taken from
    https://github.com/tomrunia/OpticalFlow_Visualization/blob/master/flow_vis/flow_vis.py'''
    @staticmethod
    def make_colorwheel():
        """
        Generates a color wheel for optical flow visualization as presented in:
            Baker et al. "A Database and Evaluation Methodology for Optical Flow" (ICCV, 2007)
            URL: http://vision.middlebury.edu/flow/flowEval-iccv07.pdf
        Code follows the original C++ source code of Daniel Scharstein.
        Code follows the the Matlab source code of Deqing Sun.
        Returns:
            np.ndarray: Color wheel
        """

        RY = 15
        YG = 6
        GC = 4
        CB = 11
        BM = 13
        MR = 6

        ncols = RY + YG + GC + CB + BM + MR
        colorwheel = np.zeros((ncols, 3))
        col = 0

        # RY
        colorwheel[0:RY, 0] = 255
        colorwheel[0:RY, 1] = np.floor(255 * np.arange(0, RY) / RY)
        col = col + RY
        # YG
        colorwheel[col:col + YG, 0] = 255 - np.floor(255 * np.arange(0, YG) / YG)
        colorwheel[col:col + YG, 1] = 255
        col = col + YG
        # GC
        colorwheel[col:col + GC, 1] = 255
        colorwheel[col:col + GC, 2] = np.floor(255 * np.arange(0, GC) / GC)
        col = col + GC
        # CB
        colorwheel[col:col + CB, 1] = 255 - np.floor(255 * np.arange(CB) / CB)
        colorwheel[col:col + CB, 2] = 255
        col = col + CB
        # BM
        colorwheel[col:col + BM, 2] = 255
        colorwheel[col:col + BM, 0] = np.floor(255 * np.arange(0, BM) / BM)
        col = col + BM
        # MR
        colorwheel[col:col + MR, 2] = 255 - np.floor(255 * np.arange(MR) / MR)
        colorwheel[col:col + MR, 0] = 255
        return colorwheel

    def flow_uv_to_colors(self, u, v, convert_to_bgr=False):
        """
        Applies the flow color wheel to (possibly clipped) flow components u and v.
        According to the C++ source code of Daniel Scharstein
        According to the Matlab source code of Deqing Sun
        Args:
            u (np.ndarray): Input horizontal flow of shape [H,W]
            v (np.ndarray): Input vertical flow of shape [H,W]
            convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.
        Returns:
            np.ndarray: Flow visualization image of shape [H,W,3]
        """
        flow_image = np.zeros((u.shape[0], u.shape[1], 3), np.uint8)
        ncols = self.colorwheel.shape[0]
        rad = np.sqrt(np.square(u) + np.square(v))
        a = np.arctan2(-v, -u) / np.pi
        fk = (a + 1) / 2 * (ncols - 1)
        k0 = np.floor(fk).astype(int)
        k1 = k0 + 1
        k1[k1 == ncols] = 0
        f = fk - k0
        for i in range(self.colorwheel.shape[1]):
            tmp = self.colorwheel[:, i]
            col0 = tmp[k0] / 255.0
            col1 = tmp[k1] / 255.0
            col = (1 - f) * col0 + f * col1
            idx = (rad <= 1)
            col[idx] = 1 - rad[idx] * (1 - col[idx])
            col[~idx] = col[~idx] * 0.75  # out of range
            # Note the 2-i => BGR instead of RGB
            ch_idx = 2 - i if convert_to_bgr else i
            flow_image[:, :, ch_idx] = np.floor(255 * col)
        return flow_image

    def flow_to_color(self, flow_uv, clip_flow=None, convert_to_bgr=False):
        """
        Expects a two dimensional flow image of shape.
        Args:
            flow_uv (np.ndarray): Flow UV image of shape [H,W,2]
            clip_flow (float, optional): Clip maximum of flow values. Defaults to None.
            convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.
        Returns:
            np.ndarray: Flow visualization image of shape [H,W,3]
        """
        assert flow_uv.ndim == 3, 'input flow must have three dimensions'
        assert flow_uv.shape[2] == 2, 'input flow must have shape [H,W,2]'
        if clip_flow is not None:
            flow_uv = np.clip(flow_uv, 0, clip_flow)
        u = flow_uv[:, :, 0]
        v = flow_uv[:, :, 1]
        rad = np.sqrt(np.square(u) + np.square(v))
        rad_max = np.max(rad)
        epsilon = 1e-5
        u = u / (rad_max + epsilon)
        v = v / (rad_max + epsilon)
        return self.flow_uv_to_colors(u, v, convert_to_bgr)




# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Unpacks a rosbag into its topics and messages
Uses the topic types to interpret the messages from each topic, 
yielding dicts for each topic containing iterables for each field.
By default unpacks all topics, but you can use any of the following keyword 
params to limit which topics are intepretted:
    - 'listTopics' = True - no unpacking - just returns a list of the topics contained in 
       the file and their associated types
    - 'importTopics' = <list of strings> - only imports the listed topics
    - 'importTypes' = <list of strings> - only imports the listed types

Message types supported are strictly those listed in the initial imports section
of this file. There are a selection of standard message types and a couple
related to event-based sensors. 
The method of importation is honed to the particular needs of
the author, sometimes ignoring certain fields, grouping data in particular ways 
etc. However it should serve as a model for anyone who wishes to import rosbags.
Although it's possible to import messages programmatically given only the 
message definition files, we have chosen not to do this, because if we did it
we would anyway need to take the resulting data and pick out the bits we wanted. 

"""

#%%

from struct import unpack
from struct import error as structError
from tqdm import tqdm

# Local imports

from .messageTypes.common import unpackHeader

from .messageTypes.dvs_msgs_EventArray import importTopic as import_dvs_msgs_EventArray
from .messageTypes.esim_msgs_OpticFlow import importTopic as import_esim_msgs_OpticFlow
from .messageTypes.geometry_msgs_PoseStamped import importTopic as import_geometry_msgs_PoseStamped
from .messageTypes.geometry_msgs_Transform import importTopic as import_geometry_msgs_Transform
from .messageTypes.geometry_msgs_TransformStamped import importTopic as import_geometry_msgs_TransformStamped
from .messageTypes.geometry_msgs_TwistStamped import importTopic as import_geometry_msgs_TwistStamped
from .messageTypes.sensor_msgs_CameraInfo import importTopic as import_sensor_msgs_CameraInfo
from .messageTypes.sensor_msgs_Image import importTopic as import_sensor_msgs_Image
from .messageTypes.sensor_msgs_Imu import importTopic as import_sensor_msgs_Imu
from .messageTypes.sensor_msgs_PointCloud2 import importTopic as import_sensor_msgs_PointCloud2
from .messageTypes.tf_tfMessage import importTopic as import_tf_tfMessage

import logging
    
def importTopic(topic, **kwargs):
    msgs = topic['msgs']
    topicType = topic['type'].replace('/','_')
    if topicType == 'dvs_msgs_EventArray': topicDict = import_dvs_msgs_EventArray(msgs, **kwargs)
    elif topicType == 'esim_msgs_OpticFlow': topicDict = import_esim_msgs_OpticFlow(msgs, **kwargs)
    elif topicType == 'geometry_msgs_PoseStamped': topicDict = import_geometry_msgs_PoseStamped(msgs, **kwargs)
    elif topicType == 'geometry_msgs_Transform': topicDict = import_geometry_msgs_Transform(msgs, **kwargs)
    elif topicType == 'geometry_msgs_TransformStamped': topicDict = import_geometry_msgs_TransformStamped(msgs, **kwargs)
    elif topicType == 'geometry_msgs_TwistStamped': topicDict = import_geometry_msgs_TwistStamped(msgs, **kwargs)
    elif topicType == 'sensor_msgs_CameraInfo': topicDict = import_sensor_msgs_CameraInfo(msgs, **kwargs)
    elif topicType == 'sensor_msgs_Image': topicDict = import_sensor_msgs_Image(msgs, **kwargs)
    elif topicType == 'sensor_msgs_Imu': topicDict = import_sensor_msgs_Imu(msgs, **kwargs)
    elif topicType == 'sensor_msgs_PointCloud2': topicDict = import_sensor_msgs_PointCloud2(msgs, **kwargs)
    elif topicType == 'tf_tfMessage': topicDict = import_tf_tfMessage(msgs, **kwargs)
    else: 
        return None
    if topicDict:
        topicDict['rosbagType'] = topic['type']
    return topicDict

def readFile(filePathOrName):
    global disable_bar
    logging.info('Attempting to import ' + filePathOrName + ' as a rosbag 2.0 file.')
    with open(filePathOrName, 'rb') as file:
        # File format string
        fileFormatString = file.readline().decode("utf-8")
        logging.info('ROSBAG file format: ' + fileFormatString)
        if fileFormatString != '#ROSBAG V2.0\n':
            logging.error('This file format (%s) might not be supported' % fileFormatString)
        eof = False
        conns = []
        chunks = []
        while not eof:
            # Read a record header
            try:
                headerLen = unpack('=l', file.read(4))[0]
            except structError:
                if len(file.read(1)) == 0: # Distinguish EOF from other struct errors 
                   # a struct error could also occur if the data is downloaded by one os and read by another.
                   eof = True
                   continue
            # unpack the header into fields 
            headerBytes = file.read(headerLen)
            fields = unpackHeader(headerLen, headerBytes)
            # Read the record data
            dataLen = unpack('=l', file.read(4))[0]
            data = file.read(dataLen)
            # The op code tells us what to do with the record
            op = unpack('=b', fields['op'])[0]
            fields['op'] = op
            if op == 2:
                # It's a message
                # AFAIK these are not found unpacked in the file
                #fields['data'] = data 
                #msgs.append(fields)
                pass
            elif op == 3:
                # It's a bag header - use this to do progress bar for the read
                chunkCount = unpack('=l', fields['chunk_count'])[0]
                pbar = tqdm(total=chunkCount, position=0, leave=True, disable=disable_bar)
            elif op == 4:
                # It's an index - this is used to index the previous chunk
                conn = unpack('=l', fields['conn'])[0]
                count = unpack('=l', fields['count'])[0]
                for idx in range(count):
                    time, offset = unpack('=ql', data[idx*12:idx*12+12])
                    chunks[-1]['ids'].append((conn, time, offset))
            elif op == 5:
                # It's a chunk
                fields['data'] = data
                fields['ids'] = []
                chunks.append(fields)
                pbar.update(len(chunks))
            elif op == 6:
                # It's a chunk-info - seems to be redundant
                pass
            elif op == 7:
                # It's a conn
                # interpret data as a string containing the connection header
                connFields = unpackHeader(dataLen, data)
                connFields.update(fields) 
                connFields['conn'] = unpack('=l', connFields['conn'])[0]
                connFields['topic'] = connFields['topic'].decode("utf-8")
                connFields['type'] = connFields['type'].decode("utf-8").replace('/', '_')
                conns.append(connFields)
    return conns, chunks

#%% Break chunks into msgs

def breakChunksIntoMsgs(chunks):
    global disable_bar
    msgs = [] 
    logging.debug('Breaking chunks into msgs ...')           
    for chunk in tqdm(chunks, position=0, leave=True, disable=disable_bar):
        for idx in chunk['ids']:
            ptr = idx[2]
            headerLen = unpack('=l', chunk['data'][ptr:ptr+4])[0]
            ptr += 4
            # unpack the header into fields 
            headerBytes = chunk['data'][ptr:ptr+headerLen]
            ptr += headerLen
            fields = unpackHeader(headerLen, headerBytes)
            # Read the record data
            dataLen = unpack('=l', chunk['data'][ptr:ptr+4])[0]
            ptr += 4
            fields['data'] = chunk['data'][ptr:ptr+dataLen]
            fields['conn'] = unpack('=l', fields['conn'])[0]
            msgs.append(fields)
    return msgs

def rekeyConnsByTopic(connDict):
    topics = {}
    for conn in connDict:
        topics[connDict[conn]['topic']] = connDict[conn]
    return topics


def importRosbag(filePathOrName, **kwargs):
    global disable_bar
    disable_bar = kwargs.get('disable_bar')
    
    logging.info('Importing file: ' + filePathOrName) 
    conns, chunks = readFile(filePathOrName)
    # Restructure conns as a dictionary keyed by conn number
    connDict = {}
    for conn in conns:
        connDict[conn['conn']] = conn
        conn['msgs'] = []
    if kwargs.get('listTopics', False):
        topics = rekeyConnsByTopic(connDict)
        logging.info('Topics in the file are (with types):')
        for topicKey, topic in topics.items():
            del topic['conn']
            del topic['md5sum']
            del topic['msgs']
            del topic['op']
            del topic['topic']
            topic['message_definition'] = topic['message_definition'].decode("utf-8")
            logging.info('    ' + topicKey + ' --- ' + topic['type'])
        return topics
    msgs = breakChunksIntoMsgs(chunks)
    for msg in msgs:     
        connDict[msg['conn']]['msgs'].append(msg)
    topics = rekeyConnsByTopic(connDict)

    importedTopics = {}
    importTopics = kwargs.get('importTopics')
    importTypes = kwargs.get('importTypes')
    if importTopics is not None:
        for topicToImport in importTopics:
            for topicInFile in list(topics.keys()):
                if topicInFile == topicToImport:
                    importedTopic = importTopic(topics[topicInFile], **kwargs)
                    if importedTopic is not None:
                        importedTopics[topicToImport] = importedTopic
                        del topics[topicInFile]            
    elif importTypes is not None:
        for typeToImport in importTypes:
            typeToImport = typeToImport.replace('/', '_')
            for topicInFile in list(topics.keys()):
                if topics[topicInFile]['type'].replace('/', '_') == typeToImport:
                    importedTopic = importTopic(topics[topicInFile], **kwargs)
                    if importedTopic is not None:
                        importedTopics[topicInFile] = importedTopic
                        del topics[topicInFile]    
    else: # import everything
        for topicInFile in list(topics.keys()):
            importedTopic = importTopic(topics[topicInFile], **kwargs)
            if importedTopic is not None:
                importedTopics[topicInFile] = importedTopic
                del topics[topicInFile]

    logging.info('')
    if importedTopics:
        logging.info('Topics imported are:')
        for topic in importedTopics.keys():
            logging.info(topic + ' --- ' + importedTopics[topic]['rosbagType'])
            #del importedTopics[topic]['rosbagType']
        logging.info('')

    if topics:
        logging.info('Topics not imported are:')
        for topic in topics.keys():
            logging.info(topic + ' --- ' + topics[topic]['type'])
        logging.info('')
    
    return importedTopics




# -*- coding: utf-8 -*-
'''
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotDvsContrast takes 'inDict' - a dict containing an imported ae file, 
as created by importAe, and creates a series of green/red plots of
polarity data.
It creates a completely separate plot for each channel which contains 'pol' data
Creates an image from events with contrast, by accumulating events up and down 
to a maximum level at which full color is used. 
The number of subplots is given by the numPlots parameter.
'distributeBy' can either be 'time' or 'events', to decide how the points 
around which data is rendered are chosen.
The events are then recruited by the time points, spreading out until
either they are about to overlap with a neighbouring point, or until 
a certain ratio of a full array is reached. 

Parameters which can be used:
 - numPlots
 - distributeBy
 - minTime, startTime
 - maxTime, endTime, stopTime
 - proportionOfPixels
 - contrast
 - flipVertical
 - flipHorizontal
 - transpose
'''

import numpy as np
import matplotlib.pyplot as plt
from math import log10, floor

from .split import selectByRange


def roundToSf(x, sig=3):
    try:
        return round(x, sig - int(floor(log10(abs(x)))) - 1)
    except ValueError:  # log of zero
        return 0


'''
nomenclature:
    idx = index
    ids = indices
'''


def idsEventsInTimeRange(events, **kwargs):
    startTime = kwargs.get('startTime', kwargs.get('minTime', kwargs.get('beginTime', events['ts'][0])))
    endTime = kwargs.get('stopTime', kwargs.get('maxTime', kwargs.get('endTime', events['ts'][-1])))
    # The following returns logical indices
    # return (events['ts'] >= startTime) & (events['ts'] < endTime)
    # Alternatively, search for the start and end indices, then return a range
    # This might be faster, given that the ts array is already sorted
    startIdx = np.searchsorted(events['ts'], startTime)
    endIdx = np.searchsorted(events['ts'], endTime)
    return range(startIdx, endIdx)


def getEventsInTimeRange(events, **kwargs):
    ids = kwargs.get('ids', idsEventsInTimeRange(events, **kwargs))
    return {
        'y': events['y'][ids],
        'x': events['x'][ids],
        'pol': events['pol'][ids],
        'ts': events['ts'][ids]
    }


def getEventImage(events, **kwargs):
    # dims might be in the events dict, but allow override from kwargs
    try:
        dimX = kwargs.get('dimX', events.get('dimX', np.max(events['x']) + 1))
        dimY = kwargs.get('dimY', events.get('dimY', np.max(events['y']) + 1))
    except ValueError:  # no defined dims and events arrays are empty
        dimX = 1
        dimY = 1
    try:
        image_type = kwargs.get('image_type', 'count')
        if image_type == 'count':
            eventImagePos = np.histogram2d(events['y'][events['pol']],
                                         events['x'][events['pol']],
                                         bins=[dimY, dimX],
                                         range=[[-0.5, dimY-0.5], [-0.5, dimX-0.5]]
                                         )[0]
            eventImageNeg = np.histogram2d(events['y'][~events['pol']],
                                         events['x'][~events['pol']],
                                         bins=[dimY, dimX],
                                         range=[[-0.5, dimY-0.5], [-0.5, dimX-0.5]]
                                         )[0]
            if kwargs.get('pol_to_show') is None or kwargs.get('pol_to_show') == 'Both':
                eventImage = eventImagePos - eventImageNeg
            elif kwargs.get('pol_to_show') == 'Pos':
                eventImage = eventImagePos
            elif kwargs.get('pol_to_show') == 'Neg':
                eventImage = - eventImageNeg
        elif image_type == 'not_polarized':
            eventImage = np.histogram2d(events['y'],
                                        events['x'],
                                        bins=[dimY, dimX],
                                        range=[[-0.5, dimY-0.5], [-0.5, dimX-0.5]]
                                        )[0]
        elif image_type == 'time_image':
            eventImage = np.zeros((dimY, dimX))
            eventImage[events['y'], events['x']] = (events['ts'] - events['ts'][0])
            eventImage = eventImage / eventImage.max()
        elif image_type == 'binary':
            eventImage = np.zeros((dimY, dimX))
            if kwargs.get('pol_to_show') is None or kwargs.get('pol_to_show') == 'Both':
                eventImage[events['y'], events['x']] = (events['pol'].astype(int) * 2 - 1)
            elif kwargs.get('pol_to_show') == 'Pos':
                eventImage[events['y'][events['pol']], events['x'][events['pol']]] = 1
            elif kwargs.get('pol_to_show') == 'Neg':
                eventImage[events['y'][~events['pol']], events['x'][~events['pol']]] = -1
        elif image_type == 'coloured':
            pos_colour = [255, 0, 0]
            neg_colour = [0, 0, 255]
            eventImage = np.full((dimY, dimX, 3), 255, dtype=np.uint8)
            if kwargs.get('pol_to_show') is None or kwargs.get('pol_to_show') == 'Both':
                eventImage[events['y'], events['x']] = [pos_colour if p else neg_colour for p in events['pol']]
            elif kwargs.get('pol_to_show') == 'Pos':
                eventImage[events['y'][events['pol']], events['x'][events['pol']]] = pos_colour
            elif kwargs.get('pol_to_show') == 'Neg':
                eventImage[events['y'][~events['pol']], events['x'][~events['pol']]] = neg_colour
    except ValueError:
        pass
    except IndexError:
        pass
    # Clip the values according to the contrast
    if not image_type == 'coloured':
        contrast = kwargs.get('contrast', 3)
        eventImage = np.clip(eventImage, -contrast, contrast)
    return eventImage


def getEventImageForTimeRange(events, **kwargs):
    events = getEventsInTimeRange(events, **kwargs)
    return getEventImage(events, **kwargs)


'''
This function accepts a dict of events and returns a new event dict
formed by collecting count number of events in the past/future from time ts
direction:
  -1: look back in the past;
  0: look equally to past and future;
  1: look to the future
'''
def getEventsByCount(events, **kwargs):
    direction = kwargs.get('direction', -1)  # default direction is past
    ts = kwargs.get('ts', 0)
    count = kwargs.get('count', 0)
    seedEventId = np.searchsorted(events['ts'], ts, side='right')
    if direction == 0:
        halfCount = int(count/2)
        firstEventId = max(seedEventId - halfCount, 0)
        lastEventId = min(seedEventId + halfCount, len(events['ts']) - 1)
        # startOrEndTime reports end time in this case
        startOrEndTime = events['ts'][lastEventId]
    else:
        startOrEndEventId = min(max(seedEventId + direction * count, 0), 
                                len(events['ts']) - 1) # limit ID within range
        startOrEndTime = events['ts'][startOrEndEventId]
        firstEventId = min(seedEventId, startOrEndEventId)
        lastEventId = max(seedEventId, startOrEndEventId)
    selectedEvents = selectByRange(events, firstEventId, lastEventId + 1)
    return selectedEvents, startOrEndTime


'''
This function accepts a dict of events and returns an event-image
formed by collecting count number of events in the past/future from time ts
direction:
  -1: look back in the past;
  0: look equally to past and future;
  1: look to the future
'''
def getEventImageByCount(events, **kwargs):
    selectedEvents, startOrEndTime = getEventsByCount(events, **kwargs)
    return getEventImage(selectedEvents, **kwargs), startOrEndTime


def plotDvsContrastSingle(inDict, **kwargs):
    frameFromEvents = getEventImage(inDict, **kwargs)
    if kwargs.get('transpose', False):
        frameFromEvents = np.transpose(frameFromEvents)
    if kwargs.get('flipVertical', False):
        frameFromEvents = np.flip(frameFromEvents, axis=0)
    if kwargs.get('flipHorizontal', False):
        frameFromEvents = np.flip(frameFromEvents, axis=1)
    contrast = kwargs.get('contrast', 3)
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
    if kwargs.get('polarised', (kwargs.get('polarized'), False)):
        cmap = kwargs.get('cmap', kwargs.get('colormap', 'seismic_r'))
        image = axes.imshow(frameFromEvents, cmap=cmap,
                            vmin=-contrast, vmax=contrast)
    else:
        cmap = kwargs.get('cmap', kwargs.get('colormap', 'gray'))
        image = axes.imshow(frameFromEvents, cmap=cmap,
                            vmin=0, vmax=contrast)
    axes.set_aspect('equal', adjustable='box')
    title = kwargs.get('title')
    if title is not None:
        axes.set_title(title)

    callback = kwargs.get('callback')
    if callback is not None:
        kwargs['axes'] = axes
        callback(frameFromEvents=frameFromEvents, **kwargs)
    return image


def plotDvsContrast(inDicts, **kwargs):
    if isinstance(inDicts, list):
        for inDict in inDicts:
            plotDvsContrast(inDict, **kwargs)
        return
    else:
        inDict = inDicts
    if not isinstance(inDict, dict):
        return
    if 'ts' not in inDict:
        title = kwargs.pop('title', '')
        if 'info' in inDict and isinstance(inDict, dict):
            fileName = inDict['info'].get('filePathOrName')
            if fileName is not None:
                print('plotDvsContrast was called for file ' + fileName)
                title = (title + ' ' + fileName).lstrip()
        for key in inDict.keys():
            kwargs['title'] = (title + ' ' + key).lstrip()
            plotDvsContrast(inDict[key], **kwargs)
        return
    # From this point onwards, it's a data-type container
    if 'pol' not in inDict:
        return
    # From this point onwards, it's a dvs container        

    # The proportion of an array-full of events which is shown on a plot
    proportionOfPixels = kwargs.get('proportionOfPixels', 0.1)
    # useAllData overrides the above - the windows used for the images together include all the data 
    useAllData = kwargs.get('useAllData', False)

    numPlots = kwargs.get('numPlots', 6)

    # Choice of distributing by 'time' or by 'numEvents'
    distributeBy = kwargs.get('distributeBy', 'time').lower()

    # % Distribute plots in a raster with a 3:4 ratio
    numPlotsX = int(round(np.sqrt(numPlots / 3 * 4)))
    numPlotsY = int(np.ceil(numPlots / numPlotsX))

    # TODO: if the actual sensor size is known, use this instead of the following defaults
    minX = kwargs.get('minX', inDict['x'].min())
    maxX = kwargs.get('maxX', inDict['x'].max())
    minY = kwargs.get('minY', inDict['y'].min())
    maxY = kwargs.get('maxY', inDict['y'].max())
    kwargs['minX'] = minX
    kwargs['maxX'] = maxX
    kwargs['minY'] = minY
    kwargs['maxY'] = maxY

    numPixelsInArray = (maxY + 1 - minY) * (maxX + 1 - minX)
    numEventsToSelectEachWay = int(round(numPixelsInArray * proportionOfPixels / 2.0))

    # The following section results in a set of tuples of first and last event idx
    # one for each plot. It does this considering the choices of distributeBy, useAllData, and proportionOfPixels

    # unpack ts for brevity
    ts = inDict['ts']

    minTime = kwargs.get('minTime', kwargs.get('startTime', kwargs.get('beginTime', ts.min())))
    maxTime = kwargs.get('maxTime', kwargs.get('stopTime', kwargs.get('endTime', ts.max())))
    minEventIdx = np.searchsorted(ts, minTime)
    maxEventIdx = np.searchsorted(ts, maxTime)
    numEvents = maxEventIdx - minEventIdx
    if distributeBy == 'time':
        totalTime = maxTime - minTime
        timeStep = totalTime / numPlots
        if useAllData:
            timeBoundaries = np.arange(minTime, maxTime + timeStep / 2, timeStep)
            firstEventIds = [
                np.where(ts >= timeBoundary)[0][0]
                for timeBoundary in timeBoundaries]
            lastEventIds = [firstEventIdx - 1 for firstEventIdx in firstEventIds]
            lastEventIds = lastEventIds[1:]
            firstEventIds = firstEventIds[:-1]
        else:
            timeCentres = np.arange(minTime + timeStep * 0.5, maxTime, timeStep)
            centreEventIds = np.searchsorted(ts, timeCentres)
            firstEventIds = [idx - numEventsToSelectEachWay for idx in centreEventIds]
            lastEventIds = [idx + numEventsToSelectEachWay for idx in centreEventIds]
    else:  # distribute by event number
        eventsPerStep = int(numEvents / numPlots)
        if useAllData:
            firstEventIds = range(minEventIdx, maxEventIdx, eventsPerStep)
            lastEventIds = [firstEventIdx - 1 for firstEventIdx in firstEventIds]
            lastEventIds = lastEventIds[1:]
            firstEventIds = firstEventIds[:-1]
        else:
            centreEventIds = range(int(eventsPerStep / 2), numEvents, eventsPerStep)
            firstEventIds = [idx - numEventsToSelectEachWay for idx in centreEventIds]
            lastEventIds = [idx + numEventsToSelectEachWay for idx in centreEventIds]
    firstEventIds = np.clip(firstEventIds, minEventIdx, maxEventIdx - 1)
    lastEventIds = np.clip(lastEventIds, minEventIdx, maxEventIdx - 1)
    firstTimes = ts[firstEventIds]
    lastTimes = ts[lastEventIds]
    timeCentres = (firstTimes + lastTimes) / 2
    titles = [
        str(roundToSf(firstTime)) + ' - ' + str(roundToSf(lastTime)) + ' s'
        for firstTime, lastTime in zip(firstTimes, lastTimes)]

    fig, allAxes = plt.subplots(numPlotsY, numPlotsX)
    if numPlots == 1:
        allAxes = [allAxes]
    else:
        allAxes = allAxes.flatten()
    fig.suptitle(kwargs.get('title', ''))

    for axes, firstEventIdx, lastEventIdx, title in zip(allAxes, firstEventIds, lastEventIds, titles):
        dvsDataDict = {
            'x': inDict['x'][firstEventIdx:lastEventIdx],
            'y': inDict['y'][firstEventIdx:lastEventIdx],
            'pol': inDict['pol'][firstEventIdx:lastEventIdx],
            'ts': inDict['ts'][firstEventIdx:lastEventIdx]
        }
        kwargs['title'] = title
        image = plotDvsContrastSingle(inDict=dvsDataDict, axes=axes, **kwargs)
    fig.colorbar(image)
    return timeCentres



# -*- coding: utf-8 -*-
"""
Copyright (C) 2021 Event-driven Perception for Robotics
Authors: Franco Di Pietro 
        Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiseDvs creates an event frame centred at time and of size timeWindow.
The user may set the amount of contrast (concurrent events in a pixel to reach
full colour) and whether to render polarity (in which case ON and OFF events
are counted against each other).
"""

import numpy as np
from scipy import ndimage
import math
# Local imports
from ..plotDvsContrast import getEventImageForTimeRange
from .visualiserBase import Visualiser


def findNearest(array, value):
    idx = np.searchsorted(array, value)  # side="left" param is the default
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        return idx - 1
    else:
        return idx


class VisualiserSkeleton(Visualiser):
    data_type = 'skeleton'

    def set_data(self, data):
        self.__data = {}
        self.__data.update(data)

    def point_to_image(self, xS, yS, image, sktN, **kwargs):
        if xS is None:
            return image
        rad = 1  # radio for each joint point - hardcoded: could be a parameter
        if (sktN == 0):
            col = [255, 0, 0]
        elif (sktN == 1):
            col = [0, 0, 255]
        else:
            col = [0, 128, 0]
        try:
            for i in range(-rad, rad):
                for j in range(-rad, rad):
                    image[yS + i, xS + j, :] = col
        except IndexError:
            pass
        return image

    def get_frame(self, time, timeWindow, **kwargs):
        if not kwargs.get('show_skeleton'):
            return None
        data = self.__data
        kwargs['startTime'] = time - timeWindow / 2
        kwargs['stopTime'] = time + timeWindow / 2

        # firstIdx = np.searchsorted(data[skt]['ts'], time - timeWindow)
        lastIdx = np.searchsorted(data['ts'], time + timeWindow) - 1
        outData = {}
        for key in data:
            if key != 'ts' and key != 'tsOffset':
                outData[key] = [data[key][lastIdx, 0],
                                data[key][lastIdx, 1]]

        return outData

    def get_settings(self):
        # settings[data_type] = {'type': 'boolean',
        #                                     'default': True
        #                                     }
        settings = {'show_skeleton': {'type': 'boolean',
                                      'default': True},
                    'show_labels': {'type': 'boolean',
                                    'default': False
                                    }
                    }

        # ,
        # 'zoom': {'type': 'boolean',
        #          'default': False
        #          },
        # 'zoomFactor': {'type': 'range',
        #                'default': 9,
        #                'min': 4,
        #                'max': 15,
        #                'step': 1
        #                },
        # 'jointZoom': {'type': 'value_list',
        #               'default': list(
        #                   self.__data[
        #                       'gt'].keys())[0],
        #               'values': self.__data[
        #                   'gt'].keys()
        #               }
        # }

        return settings



import struct
import numpy as np

def importAerdat(**kwargs):
    filePathOrName = kwargs['filePathOrName']
    with open(filePathOrName, 'rb') as f:
        content = f.read()

    ''' Packet format'''
    packet_format = 'BHHI'  # pol = uchar, (x,y) = ushort, t = uint32
    packet_size = struct.calcsize('=' + packet_format)  # 16 + 16 + 8 + 32 bits => 2 + 2 + 1 + 4 bytes => 9 bytes
    num_events = len(content) // packet_size
    extra_bits = len(content) % packet_size

    '''Remove Extra Bits'''
    if extra_bits:
        content = content[0:-extra_bits]

    ''' Unpacking'''
    event_list = list(struct.unpack('=' + packet_format * num_events, content))

    timestamps = np.array(event_list[3:][::4], dtype=float)
    x = np.array(event_list[2:][::4])
    y = np.array(event_list[1:][::4])
    pol = np.array(event_list[::4])
    tsOffset = timestamps[0]
    timestamps -= tsOffset
    timestamps /= 1e6
    out_dict = {'info': {'tsOffset': tsOffset,
                         'filePathOrName': filePathOrName,
                         'fileFormat': 'aerdat'},
                'data': {
                    'ch0': {
                        'dvs': {
                            'ts': timestamps,
                            'x': x,
                            'y': y,
                            'pol': pol
                        }}}}
    return out_dict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Author: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotSpikeogram received "inDict", which is a container for (2D) polarity events, 
containing fields with numpy arrays:    
    ts
    x
    y
    pol
A spikeogram has a row for each address represented, where the x dim 
represents time and vertical lines represent events. 
Color represents polarity (blue vs red for 1 vs 0)
By default, all addresses will be plotted as a row, 
but this will usually be impractical. So the kwargs
    xMin
    xMax
    yMin
    yMax
are used to constrain the space represented
and minTime and maxTime are used to constrain time. 
The axes are aligned so that the coordinate points to the middle of the first
row represented by that mark. 

TODO: There are separate functions for spikeograms from skin and dvs - this
should be generalised to all address-event datatypes (including ear, dvsLbl etc)

TODO: iterative adding of lines for each event is very inefficient for large
datasets - consider using plt.scatter.
"""

#%% Plot tsPix (single pixel timestamps)

import matplotlib.pyplot as plt
import matplotlib.lines as lines
import numpy as np
from tqdm import tqdm

# TODO: This code is very inefficient for large numbers of events
# better to iterate once for each pixel
def plotSpikeogramDvs(inDict, **kwargs):
    # Boilerplate for descending container hierarchy
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotSpikeogram(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotSpikeogram was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'dvs' in channelData and len(channelData['dvs']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotSpikeogram(channelData['dvs'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no polarity data')
        return
    # Break out data arrays for cleaner code
    print('plotSpikeogram working: ' + kwargs['title'])
    x = inDict['x']
    y = inDict['y']
    ts = inDict['ts']
    pol = inDict['pol']
    minX = kwargs.get('minX', np.min(x))
    maxX = kwargs.get('maxX', np.max(x))
    minY = kwargs.get('minY', np.min(y))
    maxY = kwargs.get('maxY', np.max(y))
    minTime = kwargs.get('minTime', kwargs.get('startTime', kwargs.get('beginTime', np.min(ts))))
    maxTime = kwargs.get('maxTime', kwargs.get('stopTime', kwargs.get('endTime', np.max(ts))))
    numX = maxX - minX + 1
    numY = maxY - minY + 1
    timeRange = maxTime - minTime
    selected = np.where((x >= minX) & (x <= maxX) &
                    (y >= minY) & (y <= maxY) &
                    (ts >= minTime) & (ts <= maxTime))[0]
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    for idx in tqdm(selected):
        xPos = ts[idx]
        yPos = x[idx] + y[idx]*numX
        if pol[idx]:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='blue', lw=2, axes=axes)
        else:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='red', lw=2, axes=axes)
        axes.add_line(line)
    axes.set_xlim(minTime-timeRange*0.01, maxTime+timeRange*0.01)
    axes.set_ylim(minX + minY*numX - 1, maxX + maxY*numX + 1)
    formatString = '0'+ str(int(np.log10(999))+1) + 'd'
    numRows = numX * numY
    theRange = range(minX+minY*numX, maxX+maxY*numX, int(numRows/10)) # Let's have max 10 labels
    labels = ['x' + format(np.mod(x,numX), formatString) + ',' +
              'y' + format(int(x/numX), formatString)
              for x in theRange]
    plt.yticks(theRange, labels)
    
    callback = kwargs.get('callback')
    if callback is not None:
        callback(**kwargs)

# TODO: Generalise with above for any address-event data
def plotSpikeogramSkin(inDict, **kwargs):
    # Boilerplate for descending container hierarchy
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotSpikeogramSkin(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotSpikeogram was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'skinEvents' in channelData and len(channelData['skinEvents']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotSpikeogramSkin(channelData['skinEvents'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no skinEvents data')
        return
    # Break out data arrays for cleaner code
    print('plotSpikeogram working: ' + kwargs['title'])
    addr = inDict['taxel'] + inDict['bodyPart'] * 2**10 # TODO: find general solution to express address of different datatypes
    ts = inDict['ts']
    pol = inDict['pol']
    # TODO: CropSpace on kwargs min/maxAddr
    minAddr = kwargs.get('minAddr', np.min(addr))
    maxAddr = kwargs.get('maxAddr', np.max(addr))
    minTime = kwargs.get('minTime', kwargs.get('startTime', kwargs.get('beginTime', np.min(ts))))
    maxTime = kwargs.get('maxTime', kwargs.get('stopTime', kwargs.get('endTime', np.max(ts))))
    timeRange = maxTime - minTime
    selected = np.where((ts >= minTime) & (ts <= maxTime))[0]
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    for idx in tqdm(selected):
        xPos = ts[idx]
        yPos = addr[idx]
        if pol[idx]:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='blue', lw=2, axes=axes)
        else:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='red', lw=2, axes=axes)
        axes.add_line(line)
    axes.set_xlim(minTime-timeRange*0.01, maxTime+timeRange*0.01)
    axes.set_ylim(minAddr - 1, maxAddr + 1)
    formatString = '0'+ str(int(np.log10(999))+1) + 'd'
    numRows = maxAddr - minAddr + 1
    theRange = range(minAddr, maxAddr, int(np.ceil(numRows/10))) # Let's have max 10 labels
    labels = ['addr' + format(addr, formatString) for addr in theRange]
    plt.yticks(theRange, labels)
    
    callback = kwargs.get('callback')
    if callback is not None:
        callback(**kwargs)

# TODO: Generalise with above for any address-event data
def plotSpikeogramEar(inDict, **kwargs):
    # Boilerplate for descending container hierarchy
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotSpikeogramEar(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotSpikeogram was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'ear' in channelData and len(channelData['ear']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotSpikeogramEar(channelData['ear'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no ear data')
        return
    # Break out data arrays for cleaner code
    print('plotSpikeogram working: ' + kwargs['title'])
    addr = inDict['freq'] + inDict['ch'] * 2**5 # TODO: find general solution to express address of different datatypes
    ts = inDict['ts']
    pol = inDict['pol']
    # TODO: CropSpace on kwargs min/maxAddr
    minAddr = kwargs.get('minAddr', np.min(addr))
    maxAddr = kwargs.get('maxAddr', np.max(addr))
    minTime = kwargs.get('minTime', kwargs.get('startTime', kwargs.get('beginTime', np.min(ts))))
    maxTime = kwargs.get('maxTime', kwargs.get('stopTime', kwargs.get('endTime', np.max(ts))))
    timeRange = maxTime - minTime
    selected = np.where((ts >= minTime) & (ts <= maxTime))[0]
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    for idx in tqdm(selected):
        xPos = ts[idx]
        yPos = addr[idx]
        if pol[idx]:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='blue', lw=2, axes=axes)
        else:
            line = lines.Line2D([xPos, xPos], [yPos - 0.5, yPos + 0.5], 
                                color='red', lw=2, axes=axes)
        axes.add_line(line)
    axes.set_xlim(minTime-timeRange*0.01, maxTime+timeRange*0.01)
    axes.set_ylim(minAddr - 1, maxAddr + 1)
    formatString = '0'+ str(int(np.log10(999))+1) + 'd'
    numRows = maxAddr - minAddr + 1
    theRange = range(minAddr, maxAddr, int(np.ceil(numRows/10))) # Let's have max 10 labels
    labels = ['addr' + format(addr, formatString) for addr in theRange]
    plt.yticks(theRange, labels)
    
    callback = kwargs.get('callback')
    if callback is not None:
        callback(**kwargs)
        
#%% Genericised function
        
'''
Input is a single dataType dict, with a 'ts' and an 'addr' field.
If there is a 'pol' field then do the polarised plot
'addr' is expected to be an integer.
TODO: optionally, there can be an array of labels for the addresses - 
we can use this to label spike data from any source.
Following pyNavis, we use a scatterplot instead of plotting lines for each spike.

'''

def plotSpikeogram(inDict, **kwargs):
    # TODO: could add boilerplate to descent through dict hierarchy
    # Here we a ssume a single datatype dict
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes # This line only relevant for callback
    if 'pol' in inDict:
        on = inDict['pol']
        if np.any(on):
            ff = inDict['ts'][on]
            gg = inDict['addr'][on].astype(float)
            axes.scatter(ff, gg, color='b')
        off = ~on
        if np.any(off):
            axes.scatter(inDict['ts'][off], inDict['addr'][off].astype(float), color='r')
    else:
        axes.scatter(inDict['ts'], inDict['addr'])
    return axes





# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importAe is a function for importing timestamped address-event data, given a path 
(defaulting to the current directory, intended for the yarp format) or a file.  
If the file format is not stated, there is an attempt to determine this from the file.
Then a sub-function is called, specialised for importing the data contained 
into the workspace. Depending on the format, additional data may also be imported,
including frame data, imu samples, skin data, 2d or 3d coords etc etc.
  
The output is a dictionary containing:
    - info
    - data
The exact contents varies according to the file type import but in general:
    info: this is a dict which starts life as the kwargs passed in, and is 
    augmented and modified by the actual contents of the file. It may include 
    any informational fields in file headers. Minimally, it will contain:
        - filePathAndName
        - fileFormat
    data: this is a list of dicts, one for each sensor or "channel" which has 
    been imported. Bear in mind that sub-functions may optionally split or join 
    channels. Within each dict, there is a field for each type of data contained.
    A file for example may contain data from a several sensors, but a single sensor 
    may produce polarity events ("pol"), aps samples ("aps"), imu samples etc.
    Within each of these fields, there is a dict, generally containing fields for 
    each data column, so in the case of pol events, there are 4-5 fields:
        - ts
        - x
        - y 
        - pol
        - optionally ch (channel)
        each containing a numpy array of the appropriate type for the data 
        contained, where all these arrays will be of the same length.

Aim is to support:

YARP .log - ATIS Gen1 - 24 bit - includes, IMU, Vicon, (also SKIN?)
rpg_dvs_ros - DVS/DAVIS
Maybe others? 
jAER / cAER .aedat (v1/2/3) DVS / DAVIS / Cochlea?
Samsung Gen3 VGA?
Celex ...???
"""

#%%

import os

# local imports
from .timestamps import rezeroTimestampsForImportedDicts

def getOrInsertDefault(inDict, arg, default):
    # get an arg from a dict.
    # If the the dict doesn't contain the arg, return the default, 
    # and also insert the default into the dict
    value = inDict.get(arg, default)
    if value == default:
        inDict[arg] = default
    return value


def importAe(**kwargs):
    print(kwargs)
    filePathOrName = getOrInsertDefault(kwargs, 'filePathOrName', '.')
    print(kwargs)
    if not os.path.exists(filePathOrName):
        raise FileNotFoundError("File or folder not found.")
    fileFormat = kwargs.get('fileFormat', '').lower()
    if not fileFormat:
        # Try to determine the file format
        if os.path.isdir(filePathOrName):
            # It's a path - it could contain yarp .log or frames
            listDir = os.listdir(filePathOrName)
            fileTypes = [subName.split(".")[-1] for subName in listDir]
            mostCommonFileType = max(set(fileTypes), key=fileTypes.count)
            if mostCommonFileType == 'log':
                kwargs['fileFormat'] = 'iityarp'
            elif mostCommonFileType == 'png':
                kwargs['fileFormat'] = 'frames'
            else:
                # recurse into this folder
                resultsList = []
                for subName in listDir:
                    kwargs['filePathOrName'] = os.path.join(filePathOrName, subName)
                    try:
                        result = importAe(**kwargs)
                    except ValueError:
                        continue
                    if isinstance(result, list):
                        resultsList = resultsList + result
                    else:
                        resultsList.append(result)
                if len(resultsList) > 1 and \
                    kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
                        # Optional: start the timestamps at zero for the first event
                        # This is done collectively for all the concurrent imports
                        rezeroTimestampsForImportedDicts(resultsList)
                elif len(resultsList) == 1:
                    resultsList = resultsList[0]
 
                return resultsList
        else:
            # Guess the file format based on file extension
            ext = os.path.splitext(filePathOrName)[1]
            if ext == '.dat' or ext == '.raw':
                kwargs['fileFormat'] = 'dat'
            elif ext == '.bag':
                kwargs['fileFormat'] = 'rosbag'
            elif ext == '.bin':
                kwargs['fileFormat'] = 'secdvs'
            elif ext == '.npy':
                kwargs['fileFormat'] = 'iitnpy'
            elif ext == '.aer2':
                kwargs['fileFormat'] = 'aer2'
            elif ext == '.hdf5':
                kwargs['fileFormat'] = 'hdf5'
            elif ext == '.log':
                kwargs['fileFormat'] = 'iit'
            elif ext == '.aerdat':
                kwargs['fileFormat'] = 'aerdat'
            elif ext == '.es':
                kwargs['fileFormat'] = 'es'
            # etc ...
            else:
                raise ValueError("The file format cannot be determined.")
    # Let the fileformat parameter dictate the file or folder format
    fileFormat = kwargs.get('fileFormat').lower()
    if fileFormat in ['iityarp', 'yarp', 'iit', 'log', 'yarplog']:
        if not os.path.isdir(kwargs['filePathOrName']):
            kwargs['filePathOrName'] = os.path.dirname(kwargs['filePathOrName'])
        from .importIitYarp import importIitYarp
        importedData = importIitYarp(**kwargs)
    elif fileFormat in ['rpgdvsros', 'rosbag', 'rpg', 'ros', 'bag', 'rpgdvs']:
        from .importRpgDvsRos import importRpgDvsRos
        importedData = importRpgDvsRos(**kwargs)
    elif fileFormat in ['iitnpy', 'npy', 'numpy']:
        try:
            from .importIitNumpy import importIitNumpy
            importedData = importIitNumpy(**kwargs)
        except ValueError:
            from .importInivationNumpy import importInivationNumpy
            importedData = importInivationNumpy(**kwargs)
    elif fileFormat in ['dat', 'raw']:
        from .importProph import importProph
        importedData = importProph(**kwargs)
    elif fileFormat in ['secdvs', 'bin', 'samsung', 'sec', 'gen3']:
        from .importSecDvs import importSecDvs
        importedData = importSecDvs(**kwargs)
    elif fileFormat in ['aer2']:
        from .importAer2 import importAer2
        importedData = importAer2(**kwargs)
    elif fileFormat in ['frame', 'frames', 'png', 'pngfolder', 'imagefolder']:
        from .importFrames import importFrames
        importedData = importFrames(**kwargs)
    elif fileFormat in ['hdf5', 'bimveehdf5']:
        from .importHdf5 import importHdf5
        importedData = importHdf5(**kwargs)
    elif fileFormat in ['aerdat']:
        from .importAerdat import importAerdat
        importedData = importAerdat(**kwargs)
    elif fileFormat in ['es']:
        from .importEs import importEs
        importedData = importEs(**kwargs)
    else:
        raise Exception("fileFormat: " + str(fileFormat) + " not supported.")
    #celex
    return importedData


# -*- coding: utf-8 -*-
'''
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotDvsLastTs takes a dict containing address-event data. 
Minimally, there must be x, y, and ts fields. 
For a given "time" (or a list of times, in which case subplots), 
creates a plot in which pixels are ordered according to their
last time stamp, and the order is represented by colour from red (recent)
to blue (oldest) through the spectrum.

Parameters which can be used:
 - time
 - minX
 - minY
 - maxX
 - maxY
 - flipVertical, flipHorizontal, transpose (applied in this order)
 - title
 - axis (to plot on; if not passed in, a new figure is created)
 - cmap
'''

import numpy as np
import matplotlib.pyplot as plt
from math import log10, floor
import matplotlib.colors as colors

# Optional import of rankdata method from scipy
try: 
    from scipy.stats import rankdata
except ModuleNotFoundError:
    # Stripped down version of implementation from scipy
    def rankdata(a, method='average'):
        arr = np.ravel(np.asarray(a))
        algo = 'mergesort' if method == 'ordinal' else 'quicksort'
        sorter = np.argsort(arr, kind=algo)
        inv = np.empty(sorter.size, dtype=intp)
        inv[sorter] = np.arange(sorter.size, dtype=intp)
        if method == 'ordinal':
            return inv + 1
        arr = arr[sorter]
        obs = np.r_[True, arr[1:] != arr[:-1]]
        dense = obs.cumsum()[inv]
        # cumulative counts of each unique value
        count = np.r_[np.nonzero(obs)[0], len(obs)]
        # average method
        return .5 * (count[dense] + count[dense - 1] + 1)

def roundToSf(x, sig=3): # https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
    try:
        return round(x, sig-int(floor(log10(abs(x))))-1)
    except ValueError: # log of zero
        return 0

def plotDvsLastTsSingle(inDict, **kwargs):
    
    time = kwargs.get('time', kwargs.get('maxTime', kwargs.get('lastTime', 
           kwargs.get('ts', kwargs.get('maxTs', kwargs.get('lastTs', 
           np.max(inDict['ts'])))))))
    minTime = kwargs.get('minTime', kwargs.get('firstTime', 
              kwargs.get('minTs', kwargs.get('firstTs', 
              np.min(inDict['ts'])))))
        
    # TODO: if the actual sensor size is known, use this instead of the following 
    minY = kwargs.get('minY',inDict['y'].min())
    maxY = kwargs.get('maxY',inDict['y'].max())
    minX = kwargs.get('minX',inDict['x'].min())
    maxX = kwargs.get('maxX',inDict['x'].max())
    sizeX = maxX - minX + 1
    sizeY = maxY - minY + 1
    tsArray = np.ones((sizeY, sizeX), dtype=float) * -1

    # populate the array by running time forward to time
    chosenLogical = inDict['ts'] <= time
    chosenLogical &= inDict['ts'] >= minTime
    chosenLogical &= inDict['x'] >= minX
    chosenLogical &= inDict['x'] <= maxX
    chosenLogical &= inDict['y'] >= minY
    chosenLogical &= inDict['y'] <= maxY
    xChosen = inDict['x'][chosenLogical] - minX
    yChosen = inDict['y'][chosenLogical] - minY
    tsChosen = inDict['ts'][chosenLogical]
    for x, y, ts in zip(xChosen, yChosen, tsChosen):
        tsArray[y, x] = ts         
    
    ordinal = kwargs.get('ordinal', False)
    if ordinal:
        tsArrayFlattened = tsArray.flatten()
        tsOrdinal = rankdata(tsArrayFlattened, method='dense') - 1 # min rank is 1
        if np.any(tsArrayFlattened == -1):
            tsOrdinal -= 1 # If there are unset timestamps, they will have rank 0 - push these to -1
        tsArray = tsOrdinal.reshape(tsArray.shape)
    else:
        pass # TODO: cardinal
    axes = kwargs.get('axes')
    if axes is None:
        fig, axes = plt.subplots()
        kwargs['axes'] = axes

    transpose = kwargs.get('transpose', False)
    if transpose:
        tsArray = np.transpose(tsArray)
    cmap = plt.get_cmap(kwargs.get('cmap', 'jet'))
    cmap.set_under(color='white')
    image = axes.imshow(tsArray, cmap=cmap, norm=colors.Normalize(vmin=0, vmax=np.max(tsArray)))
    axes.set_aspect('equal', adjustable='box')
    axes.grid(b=False)
    if kwargs.get('flipVertical', False):
        axes.invert_yaxis()
    if kwargs.get('flipHorizontal', False):
        axes.invert_xaxis()
    title = kwargs.get('title')
    if title is not None:
        axes.set_title(title)
    axes.set_title(str(roundToSf(minTime)) + ' - ' + str(roundToSf(time)) + ' s')
    
    callback = kwargs.get('callback')
    if callback is not None:
        callback(tsArray=tsArray, **kwargs)
    
    return image

def plotDvsLastTs(inDict, **kwargs):
    # Boilerplate for descending higher level containers
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotDvsLastTs(inDictInst, **kwargs)
        return
    if 'info' in inDict: # Top level container
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotDvsContrast was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'dvs' in channelData and len(channelData['dvs']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotDvsLastTs(channelData['dvs'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no polarity data')
        return
    times = kwargs.get('time', kwargs.get('maxTime', kwargs.get('lastTime', 
           kwargs.get('ts', kwargs.get('maxTs', kwargs.get('lastTs', 
           np.max(inDict['ts'])))))))
    minTimes = kwargs.get('minTime', kwargs.get('firstTime', 
              kwargs.get('minTs', kwargs.get('firstTs', 
              np.min(inDict['ts'])))))
    if np.isscalar(times):
        times = [times]
        minTimes = [minTimes]
    numPlots = len(times)
    numPlotsX = int(round(np.sqrt(numPlots / 3 * 4)))
    numPlotsY = int(np.ceil(numPlots / numPlotsX))
    fig, allAxes = plt.subplots(numPlotsY, numPlotsX)
    if numPlots == 1:
        allAxes = [allAxes]
    else:
        allAxes = allAxes.flatten()
    fig.suptitle(kwargs.get('title', ''))
    for time, minTime, axes in zip(times, minTimes, allAxes):
        kwargs['time'] = time
        kwargs['minTime'] = minTime
        kwargs['axes'] = axes
        plotDvsLastTsSingle(inDict, **kwargs)



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window).
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict.
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiserBoundingBoxes doesn't actually create a visualisation but rather
queries the data for the bounding boxes within the time window and passes
these out for visualisation as an overlay.
"""

import numpy as np

# Local imports
from .visualiserBase import Visualiser


class VisualiserBoundingBoxes(Visualiser):

    data_type = 'boundingBoxes'

    def __init__(self, data):
        self.set_data(data)

    def set_data(self, data):
        self.__data = {}
        self.__data.update(data)

    def get_data(self):
        return self.__data

    def get_frame(self, time, timeWindow, **kwargs):
        if self.__data is None or not kwargs.get('show_bounding_boxes', True):
            return None
        gt_bb = self.__data
        indices = abs(gt_bb['ts'] - time) < timeWindow
        if not any(indices):
            if not kwargs.get('interpolate'):
                return None
        if kwargs.get('interpolate'):
            boxes = []
            for label in np.unique(gt_bb['label']):
                label_mask = gt_bb['label'] == label
                ts = gt_bb['ts'][label_mask]
                minY = gt_bb['minY'][label_mask]
                minX = gt_bb['minX'][label_mask]
                maxY = gt_bb['maxY'][label_mask]
                maxX = gt_bb['maxX'][label_mask]

                i1 = np.searchsorted(ts, time)
                i0 = i1 - 1
                if i0 < 0:
                    if abs(ts[0] - time) < timeWindow:
                        if kwargs.get('with_labels', True) and 'label' in gt_bb.keys():
                            boxes.append((minY[0], minX[0], maxY[0], maxX[0], label))
                        else:
                            boxes.append((minY[0], minX[0], maxY[0], maxX[0]))
                        continue
                    else:
                        continue
                if i1 >= len(ts):
                    if abs(ts[-1] - time) < timeWindow:
                        if kwargs.get('with_labels', True) and 'label' in gt_bb.keys():
                            boxes.append((minY[-1], minX[-1], maxY[-1], maxX[-1], label))
                        else:
                            boxes.append((minY[-1], minX[-1], maxY[-1], maxX[-1]))
                        continue
                    else:
                        continue

                minY_interp = minY[i0] + ((minY[i1] - minY[i0]) / abs(ts[i1] - ts[i0])) * (time - ts[i0])
                minX_interp = minX[i0] + ((minX[i1] - minX[i0]) / abs(ts[i1] - ts[i0])) * (time - ts[i0])
                maxY_interp = maxY[i0] + ((maxY[i1] - maxY[i0]) / abs(ts[i1] - ts[i0])) * (time - ts[i0])
                maxX_interp = maxX[i0] + ((maxX[i1] - maxX[i0]) / abs(ts[i1] - ts[i0])) * (time - ts[i0])
                if kwargs.get('with_labels', True) and 'label' in gt_bb.keys():
                    boxes.append((minY_interp, minX_interp, maxY_interp, maxX_interp, label))
                else:
                    boxes.append((minY_interp, minX_interp, maxY_interp, maxX_interp))
            boxes = np.array(boxes).astype(int)
        else:
            boxes = np.column_stack((gt_bb['minY'][indices], gt_bb['minX'][indices],
                                     gt_bb['maxY'][indices], gt_bb['maxX'][indices])).astype(int)
            if kwargs.get('with_labels', True) and 'label' in gt_bb.keys():
                labels = gt_bb['label'][indices].astype(int)
                boxes = np.column_stack([boxes, labels])
        boxes = np.unique(boxes, axis=0)
        return boxes

    def get_settings(self):
        settings = {'with_labels': {'type': 'boolean',
                                    'default': True
                                    },
                    'show_bounding_boxes': {'type': 'boolean',
                                            'default': True
                                            },
                    'interpolate': {'type': 'boolean',
                                    'default': False
                                    }
                    }
        return settings



# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

The interpretMessages function receives a list of messages and returns
a dict with one field for each data field in the message, where the field
will contain an appropriate iterable to contain the interpretted contents of each message.

This function imports the ros message type defined at:
http://docs.ros.org/melodic/api/geometry_msgs/html/msg/TwistStamped.html
"""

#%%

from tqdm import tqdm
import numpy as np

from .common import unpackRosString, unpackRosTimestamp, unpackRosFloat64Array

def importTopic(msgs, **kwargs):
    disable_bar = kwargs.get('disable_bar')
    sizeOfArray = 1024
    ts = np.zeros((sizeOfArray), dtype=np.float64)
    linV = np.zeros((sizeOfArray, 3), dtype=np.float64)
    angV = np.zeros((sizeOfArray, 3), dtype=np.float64)
    for idx, msg in enumerate(tqdm(msgs, position=0, leave=True, disable=disable_bar)):
        if sizeOfArray <= idx:
            ts = np.append(ts, np.zeros((sizeOfArray), dtype=np.float64))
            linV = np.concatenate((linV, np.zeros((sizeOfArray, 3), dtype=np.float64)))
            angV = np.concatenate((angV, np.zeros((sizeOfArray, 3), dtype=np.float64)))
            sizeOfArray *= 2
        data = msg['data']
        #seq = unpack('=L', data[0:4])[0]
        ts[idx], ptr = unpackRosTimestamp(data, 4)
        frame_id, ptr = unpackRosString(data, ptr)
        linV[idx, :], ptr = unpackRosFloat64Array(data, 3, ptr)
        angV[idx, :], _ = unpackRosFloat64Array(data, 3, ptr)
    # Crop arrays to number of events
    numEvents = idx + 1
    ts = ts[:numEvents]
    linV = linV[:numEvents]
    angV = angV[:numEvents]
    outDict = {
        'ts': ts,
        'linV': linV,
        'angV': angV}
    return outDict




import json
import numpy as np
import re
import os

def importSkeletonDataLog(filePath, keys):
    pattern = re.compile('\d* (\d*.\d*) SKLT \((.*)\)')
    with open(filePath) as f:
        content = f.readlines()
    data_dict = {k: [] for k in keys}
    timestamps = []
    for line in content:
        ts, data = pattern.findall(line)[0]
        data = np.array(data.split(' ')).astype(int).reshape(-1, 2)
        for d, label in zip(data, data_dict):
            data_dict[label].append(d)
        timestamps.append(ts)
    data_dict['ts'] = np.array(timestamps).astype(float)
    for d in data_dict:
        data_dict[d] = np.array(data_dict[d])
    return data_dict

def importSkeleton(**kwargs):
    with open(kwargs.get('filePathOrName'), 'r') as f:
        data_dict = json.load(f)
    return importSkeletonDataLog(os.path.join(os.path.dirname(kwargs['filePathOrName']), data_dict['file']), data_dict['labels'])



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Author: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Various functions for working with timestamps

zeroTimestamps(imports)

imports can either be an importDict as produced by importAe, or it can be a list of them. 
For each import, for each channel and for each datatype, the first timestamp is found. 
We assume that these timestamps are already synchronised.
We take the lowest, and use that to zero all the timestamps. 

unwrapTimestamps()
taking a numpy array (expected to by dtype=float), return the array where
any timestamp wrap events have been treated by advancing all the timestamps 
after every wrap event.
The 'wrapTime' argument is the time at which wrapping is expected to occur.
If this is not given, then the maximum timestamp before each wrap event is used.
    
"""

#%%

import numpy as np
import math

def zeroTimestampsForADataType(dataTypeDict, tsOffset=None):
    # Probably the common format is 'ts' for all dtypes, 
    # but handle any exceptions here, example: if dtypeName == 'frame':
    tsOffsetInitial = dataTypeDict.get('tsOffset', 0)
    if tsOffset is None:
        tsOffset = - np.min(dataTypeDict['ts'])
    dataTypeDict['ts'] = dataTypeDict['ts'] + tsOffset
    dataTypeDict['tsOffset'] = tsOffset + tsOffsetInitial

def getFirstTimestampForAChannel(channelDict):
    firstTimestamp = float(np.inf)    
    for dtypeName in channelDict:
        # Probably the common format is 'ts' for all dtypes, 
        # but handle any exceptions here, example: if dtypeName == 'frame':
        try:
            firstTimestamp = min(firstTimestamp, 
                             channelDict[dtypeName]['ts'][0]) 
        except KeyError:
            # This dataType doesn't have a ts; no problem. 
            pass
    return firstTimestamp

def zeroTimestampsForAChannelInSitu(channelDict, tsOffset=None):
    if tsOffset is None:
        tsOffset = -getFirstTimestampForAChannel(channelDict)
    for dtypeName in channelDict:
        try:
            channelDict[dtypeName]['ts'] = channelDict[dtypeName]['ts'] + tsOffset
            channelDict[dtypeName]['tsOffset'] = tsOffset
        except KeyError:
            # This dataType doesn't have a ts; no problem. 
            pass


'''
Old behaviour was in situ - the same dicts were altered;
This is retained as default behaviour for now for backwards compatibility
TODO: deprecate old behaviour
'''
def zeroTimestampsForAChannel(channelDict, tsOffset=None, inSitu=True):
    if inSitu:
        zeroTimestampsForAChannelInSitu(channelDict, tsOffset)
        return 
    if tsOffset is None:
        tsOffset = -getFirstTimestampForAChannel(channelDict)
    newChannelDict = {}
    for dataTypeName in channelDict.keys():
        newDataTypeDict = channelDict[dataTypeName].copy()
        try:
            newDataTypeDict['ts'] = newDataTypeDict['ts'] + tsOffset
            newDataTypeDict['tsOffset'] = tsOffset
        except KeyError:
            # This dataType doesn't have a ts; no problem. 
            pass
        newChannelDict[dataTypeName] = newDataTypeDict
    return newChannelDict


'''
This function is similar to rezeroTimestampsForAnImportedDict below,
but doesn't assume the info/data level of the dict structure.
Also it doesn't modify dicts in situ but rather creates new ones.
'''
def rezeroTimestampsForMultipleChannels(inDict):
    # Find largest (i.e. least negative) offset
    tsOffset = float(-np.inf)
    for channelName in inDict:
        for dataType in inDict[channelName]:
            tsOffset = max(
                tsOffset,
                inDict[channelName][dataType].get('tsOffset', -np.inf))
    # Now we have the least negative tsOffset, iterate through all, reapplying it
    outDict = {}
    for channelName in inDict:
        outDict[channelName] = {}
        for dataType in inDict[channelName]:
            outDict[channelName][dataType] = inDict[channelName][dataType].copy()
            try:
                tsOffsetCurrent = outDict[channelName][dataType].get('tsOffset', 0.0)
                outDict[channelName][dataType]['ts'] = (
                    outDict[channelName][dataType]['ts'] +
                    tsOffset - tsOffsetCurrent)
                outDict[channelName][dataType]['tsOffset'] = tsOffset
            except KeyError:
                # This dataType doesn't have a ts; no problem. 
                pass
    return outDict, tsOffset

'''
First zero timestamps for each channel individually, and then realign them all jointly
'''
def zeroTimestampsForMultipleChannels(inDict):
    intermediateDict = {}
    for channelName in inDict:
        intermediateDict[channelName] = zeroTimestampsForAChannel(
            inDict[channelName], inSitu=False)
    return rezeroTimestampsForMultipleChannels(intermediateDict)

    
''' 
This function receives a single importedDict from importing one file. 
It is focused on the problem of aligning data between multiple datatypes,
Within the dataType dicts there are timestamps which may or may not come from a device, like Stefi.
Let's call these data-level timestamps.
At the level of the info branch of the dict we may have system clock time 
at which a recording started.
Let's call these info-level timestamps.
This function only considers data-level timestamps.
It assumes that timestamps have already been aligned individually within datatypes.
Therefore each dataType dict which contains 'ts' should also contain 'tsOffset',
which says how much time was added to the 'ts' field.
This function aligns these so that the first event across the file is at ts = 0.
The tsOffset needed to achieve this is then added to the info branch of the dict
as tsOffsetFromData.
TODO: Rebuild by calling rezeroTimestampsForMultipleChannels, above; this would
also remove in-situ modification behaviour.
'''
def rezeroTimestampsForAnImportedDict(importedDict):
    # Find largest (i.e. least negative) offset
    tsOffset = float(-np.inf)
    for channelName in importedDict['data']:
        for dataType in importedDict['data'][channelName]:
            tsOffset = max(tsOffset, importedDict['data'][channelName][dataType].get('tsOffset', -np.inf))
    # Now we have the least negative tsOffset, iterate through all, reapplying it
    for channelName in importedDict['data']:
        for dataType in importedDict['data'][channelName]:
            try:
                tsOffsetCurrent = importedDict['data'][channelName][dataType].get('tsOffset', 0.0)
                importedDict['data'][channelName][dataType]['ts'] = \
                    importedDict['data'][channelName][dataType]['ts'] + tsOffset - tsOffsetCurrent
                importedDict['data'][channelName][dataType]['tsOffset'] = tsOffset
            except KeyError:
                # This dataType doesn't have a ts; no problem. 
                pass
    importedDict['info']['tsOffsetFromData'] = tsOffset

''' 
This function receives a list of importedDicts - each one from importing one file. 
It is focused on the problem of aligning data between multiple datatypes,
Within the dataType dicts there are timestamps which might come from a device -
let's call these data-level timestamps.
At the level of info branch of the dict we have system clock time at which 
a recording started - Let's call these info-level timestamps.
It first calls the above function rezeroTimestampsForAnImportedDict for each file, 
so that timestamps are aligned individually within the dict for each file. 
The timestamp which was used to achieve this will be in ['info']['tsOffsetFromData'].
There may additionally be ['info']['tsOffsetFromInfo'] - the info level timestamps.
If info/level timestamops are available for all files, these are used to rezero
all the data across multiple files.
If any are absent, then they are ignored, and alignment across multiple files
is based on data-level timestamps.
'''      
def rezeroTimestampsForImportedDicts(importedDicts):
    if not isinstance(importedDicts, list):
        importedDicts = [importedDicts]
    # Find largest (i.e. least negative) offset
    tsOffsetFromData = float(-np.inf)
    # Confusingly, info-level timestamps run in the other direction! TODO: could fix this
    tsOffsetFromInfo = float(np.inf)
    allHaveInfoTsOffset = True
    for importedDict in importedDicts:
        rezeroTimestampsForAnImportedDict(importedDict)
        tsOffsetFromData = max(tsOffsetFromData, importedDict['info']['tsOffsetFromData'])
        if importedDict['info'].get('tsOffsetFromInfo', 0) != 0: # 0 is a placeholder used by the exporter - if it's zero, we assume it is not a real value that we care about.
            tsOffsetFromInfo = min(tsOffsetFromInfo, importedDict['info']['tsOffsetFromInfo'])
        else:
            allHaveInfoTsOffset = False
    if len(importedDicts) == 1:
        return
    # Now we have the extreme tsOffsets, iterate through all, reapplying 
    for importedDict in importedDicts:
        if allHaveInfoTsOffset and len(importedDicts) > 1:
            if tsOffsetFromInfo == importedDict['info']['tsOffsetFromInfo']:
                # The reason for this if clause is to catch a floating point 
                # precision error due to mixing very big and very small timestamps. 
                tsOffset = importedDict['info']['tsOffsetFromData']                
            else:
                tsOffset = importedDict['info']['tsOffsetFromData'] - tsOffsetFromInfo + importedDict['info']['tsOffsetFromInfo']
            
        else:
            tsOffset = tsOffsetFromData
        for channelName in importedDict['data']:
            for dataType in importedDict['data'][channelName]:
                try:
                    tsOffsetCurrent = importedDict['data'][channelName][dataType].get('tsOffset', 0.0)
                    importedDict['data'][channelName][dataType]['ts'] = \
                        importedDict['data'][channelName][dataType]['ts'] + tsOffset - tsOffsetCurrent
                    importedDict['data'][channelName][dataType]['tsOffset'] = tsOffset
                except KeyError:
                    # This dataType doesn't have a ts; no problem. 
                    pass

'''
using container as a general term for datatype dicts, channels, fileDicts or any hierarchical list of them...
recurse through the structure, modifying timestamps and tsOffsets by a fixed amount
'''
def offsetTimestampsForAContainer(container, offset):
    if isinstance(container, list):
        for elem in container:
            offsetTimestampsForAContainer(elem, offset)
        return
    if isinstance(container, dict):
        if 'ts' in container:
            # It's a datatype dict
            container['ts'] = container['ts'] + offset
            container['tsOffset'] = container['tsOffset'] + offset
        else:
            for field in container.values():
                offsetTimestampsForAContainer(field, offset)
    else:
        # We have descended too far
        return

'''
Takes ts where as int or float and returns it as float64, unwrapping where necessary.
If you pass in wrapTime that takes precedence - make sure you pass in the right type though.
If not, but you pass in tsBits (the number of timestamp bits), 
then the wrapTime is 2**wrapTime.
Otherwise, tsBits is guessed by looking at the highest actual ts. 
'''

def unwrapTimestamps(ts, **kwargs):
    wrapTime = kwargs.get('wrapTime', 2**int(kwargs.get('tsBits', np.ceil(np.log2(np.max(ts)))))) # This would fail in the edge case of e.g. max ts=64, but it's highly unlikely
    # In the case that tsBits has been explicitly passed in, assume the input
    # array is uint and remove any extra bits before continuing. 
    tsBits = kwargs.get('tsBits')
    if tsBits is not None and tsBits < 32:
        ts = ts & (np.uint32(0x1 << tsBits) - 1)
    ts = ts.astype(float) 
    diff = ts[1:] - ts[:-1] 
    wrapPoints = np.where(diff < 0)[0]
    for wrapPoint in wrapPoints:
        ts[wrapPoint+1:] = ts[wrapPoint+1:] + wrapTime
    return ts

def cropSelectedFields(dataTypeDict, fieldList, selectedBool):
    for fieldName in fieldList:
        dataTypeDict[fieldName] = dataTypeDict[fieldName][selectedBool]

'''
Accepts a container at any level of the container hierarchy and finds the highest timestamp contained
'''
def getLastTimestamp(inDict):
    lastTs = 0
    if isinstance(inDict, list):
        for inDictElement in inDict:
            lastTs = max(lastTs, getLastTimestamp(inDictElement))
    elif isinstance(inDict, dict):
        if 'ts' in inDict:
            return inDict['ts'][-1]
        else: # It's a dictionary - go through it's elements
            for keyName in inDict.keys():
                lastTs = max(lastTs, getLastTimestamp(inDict[keyName]))
    return lastTs

'''
cropDataByTimeRange was here - replaced by cropTime in split.py
'''

'''
Sort a dict containing ts according to ts, applying the new sort order to all
the fields where axis 0 
'''
def sortDataTypeDictByTime(inDict):
    try:
        ts = inDict['ts']
    except KeyError:
        return inDict
    ids = np.argsort(ts) 
    numEvents = ts.shape[0]
    outDict = {}
    for fieldName in inDict.keys():
        try:
            if isinstance(inDict[fieldName], list):
                assert len(inDict[fieldName]) == numEvents
                outDict[fieldName] = [inDict[fieldName][idx] for idx in ids]
            elif isinstance(inDict[fieldName], np.ndarray):
                assert len(inDict[fieldName]) == numEvents
                outDict[fieldName] = inDict[fieldName][ids]
            else:
                outDict[fieldName] = inDict[fieldName]
        except (AssertionError, TypeError):
            outDict[fieldName] = inDict[fieldName]
    return outDict
    
'''
If two or more file-level containers are merged, in general it's not possible 
to align their timestamps without some extra prior knowledge. Here we reset
all tsOffset fields, to avoid them being used to readjust timestamps incorrectly. 
We assume that there's no overlap of channel names. 
We also assume 
'''    
def mergeContainers(listOfDicts):
    outDict = {'info': {'filePathOrName': 'merged'},
               'data': {}}
    for inDict in listOfDicts:
        outDict['data'].update(inDict['data'])
    for channelName in outDict['data']:
        for dataTypeName in outDict['data'][channelName]:
            outDict['data'][channelName][dataTypeName]['tsOffset'] = 0
    return outDict


# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def findNearest(array, value):
    idx = np.searchsorted(array, value) # side="left" param is the default
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx

def findNearestTsIdx(inDict, tsIdeal):
    return findNearest(inDict['ts'], tsIdeal)

def getNearestEvent(inDict, tsIdeal):
    idx = findNearest(inDict['ts'], tsIdeal)
    return getEvent(inDict, idx)

'''
Possibly a temporary home for this utility function
Given any dataType container dict, return a dict which just contains the data
for the indexed event or sample
'''
def getEvent(inDict, idx):
    outDict = inDict.copy()
    numEvents = len(outDict['ts'])
    for key in outDict.keys():
        try:
            if len(outDict[key]) == numEvents:
                outDict[key] = outDict[key][idx]
        except TypeError:
            continue
    return outDict



#%% LEGACY CODE - timestamps for different data types present in aedat
# There are exceptions around the timestamps for frame data to consider 
'''
def FindFirstAndLastTimeStamps(aedat):
    
    This is a sub-function of importAedat. 
    For each field in aedat['data'], it finds the first and last timestamp. 
    The min and max of these respectively are put into aedat.info
    
    
    # Clip arrays to correct size and add them to the output structure.
    # Also find first and last timeStamps
    
    if not 'data' in aedat:
        print('No data found from which to extract time stamps')
        return aedat
    
    firstTimeStamp = np.inf
    lastTimeStamp = 0
    
    if 'special' in aedat['data']:
    	if aedat['data']['special']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['special.timeStamp'][0]
    	if aedat['data']['special']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['special']['timeStamp'][-1]
    
    if 'polarity' in aedat['data']:
    	if aedat['data']['polarity']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['polarity']['timeStamp'][0]
    	if aedat['data']['polarity']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['polarity']['timeStamp'][-1]
    
    if 'frame' in aedat['data']:    
        if 'timeStampExposureStart' in aedat['data']['frame']:
            if aedat['data']['frame']['timeStampExposureStart'][0] < firstTimeStamp:
                firstTimeStamp = aedat['data']['frame']['timeStampExposureStart'][0]
            if aedat['data']['frame']['timeStampExposureEnd'][-1] > lastTimeStamp:
                lastTimeStamp = aedat['data']['frame']['timeStampExposureEnd'][-1]
        else:
            if aedat['data']['frame']['timeStampStart'][0] < firstTimeStamp:
                firstTimeStamp = aedat['data']['frame']['timeStampStart'][0]
            if aedat['data']['frame']['timeStampEnd'][-1] > lastTimeStamp:
                lastTimeStamp = aedat['data']['frame']['timeStampEnd'][-1]
    
    if 'imu6' in aedat['data']:
    	if aedat['data']['imu6']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['imu6']['timeStamp'][0]
    	if aedat['data']['imu6']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['imu6']['timeStamp'][-1]
    
    if 'sample' in aedat['data']:
    	if aedat['data']['sample']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['sample']['timeStamp'][0]
    	if aedat['data']['sample']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['sample']['timeStamp'][-1]
    
    if 'ear' in aedat['data']:
    	if aedat['data']['ear']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['ear']['timeStamp'][0]
    	if aedat['data']['ear']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['ear']['timeStamp'][-1]
    
    if 'point1D' in aedat['data']:
    	if aedat['data']['point1D']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['point1D']['timeStamp'][0]
    	if aedat['data']['point1D']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['point1D']['timeStamp'][-1]
    
    if 'point2D' in aedat['data']:
    	if aedat['data']['point2D']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['point2D']['timeStamp'][0]
    	if aedat['data']['point2D']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['point2D']['timeStamp'][-1]

    if 'point3D' in aedat['data']:
    	if aedat['data']['point3D']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['point3D']['timeStamp'][0]
    	if aedat['data']['point3D']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['point3D']['timeStamp'][-1]

    if 'point4D' in aedat['data']:
    	if aedat['data']['point4D']['timeStamp'][0] < firstTimeStamp:
    		firstTimeStamp = aedat['data']['point4D']['timeStamp'][0]
    	if aedat['data']['point4D']['timeStamp'][-1] > lastTimeStamp:
    		lastTimeStamp = aedat['data']['point4D']['timeStamp'][-1]
    
    aedat['info']['firstTimeStamp'] = firstTimeStamp
    aedat['info']['lastTimeStamp'] = lastTimeStamp

    return aedat
'''


# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
         Aiko Dinale
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Contains various functions for splitting data sets:

Two general functions in which one chosen field is used to divide a whole dataset
- selectByLabel
- splitByLabel

selectByBool: divide a whole dataset based on a boolean array which is passed in

splitByPolarity: divide events according to a boolean "pol" field (intended for dvs events)

Two functions for narrowing down on the data by time or space:
- cropTime (cropTemporal)
- cropSpace (cropSpatial)
cropTime works on ts.
cropSpace works from optional min/max|X/Y/Z parameters; 
    it has implementation for various datatypes:
    - dvs, flow: operates on x and y fields 
    - point: operates on the 3 dimensions, x, y, and z, of a 'point' field.
    - frame: crops frames down to a chosen 2D spatial window.
"""

#%%
import numpy as np
from itertools import compress
from math import fabs

# local imports
from .timestamps import rezeroTimestampsForImportedDicts, sortDataTypeDictByTime

# In selectByLabel, the field from which a value is selected already exists in the iDict
def selectByLabel(inDict, labelName, labelValue):
    selectedEvents = inDict[labelName] == labelValue
    if not np.any(selectedEvents):
        return None
    outDict = {}
    for fieldName in inDict.keys():
        if fieldName != labelName:
            try:
                assert len(inDict[fieldName]) == len(selectedEvents)
                outDict[fieldName] = inDict[fieldName][selectedEvents]
            except (AssertionError, TypeError):
                outDict[fieldName] = inDict[fieldName]
    return outDict

# In selectByBool, we pass in a boolean array the same length as ts and the 
# other iterables, and use this for selection of only those elements which are 
# true in the boolean array
def selectByBool(inDict, selectedEvents):
    if not np.any(selectedEvents):
        return None
    outDict = {}
    for fieldName in inDict.keys():
        try:
            assert len(inDict[fieldName]) == len(selectedEvents)
            if isinstance(inDict[fieldName], list):
                outDict[fieldName] = list(compress(inDict[fieldName], selectedEvents))
            else:
                outDict[fieldName] = inDict[fieldName][selectedEvents]
        except (AssertionError, TypeError): # TypeError for case that value has no len(); #AssertionError, in case it does but that len is not the same as the ts len.
            outDict[fieldName] = inDict[fieldName]
    return outDict


def selectByRange(inDict, firstIdx=0, lastIdx=None):
    if lastIdx is None:
        lastIdx = len(inDict['ts'])
    outDict = {}
    if firstIdx >= lastIdx:
        return outDict
    for fieldName in inDict.keys():
        try:
            assert len(inDict[fieldName]) == len(inDict['ts'])
            outDict[fieldName] = inDict[fieldName][firstIdx:lastIdx]
        except (AssertionError, TypeError): # TypeError for case that value has no len(); #AssertionError, in case it does but that len is not the same as the ts len.
            outDict[fieldName] = inDict[fieldName]
    return outDict


# Choose a field and split the dict into dicts each of which contain a unique
# value for that field
# if param outList is true, return a list of dicts,otherwise return a dict
# of dicts, each of which having the name of the unique field value
def splitByLabel(inDict, labelName, outList=False):
    labels = np.unique(inDict[labelName])
    if outList:
        outList = []
        for label in labels:
            selectedEvents = inDict[labelName] == label
            outDict = {}
            for fieldName in inDict.keys():
                try:
                    assert len(inDict[fieldName]) == len(selectedEvents)
                    outDict[fieldName] = inDict[fieldName][selectedEvents]
                except (TypeError, AssertionError):
                    outDict[fieldName] = inDict[fieldName]
            outList.append(outDict)
        return outList
    else:
        outDictParent = {}
        for label in labels:
            selectedEvents = inDict[labelName] == label
            outDict = {}
            for fieldName in inDict.keys():
                try:
                    assert len(inDict[fieldName]) == len(selectedEvents)
                    outDict[fieldName] = inDict[fieldName][selectedEvents]
                except (TypeError, AssertionError):
                    outDict[fieldName] = inDict[fieldName]
            outDictParent[label] = outDict
        return outDictParent

''' Intended for data where samples come out in bursts at certain timestamps,
output these bursts in separate dicts '''
def splitByTimestamp(inDict, outList=False):
    return splitByLabel(inDict, 'ts', outList=outList)
    
'''
receives a dict containing (probably) dvs events
returns a dict containing two dicts, labelled 0 and 1, for the polarities found
Although redundant, the pol field is maintained within each dict for compatibility
This is similar to splitByLabel but specialised for dvs;
it is retained because True and False values make awkward dictionary keys,
so here they are replaced by strings '0' and '1'
'''
def splitByPolarity(inDict):
    outDict = {
        '0': {},
        '1': {} }
    for key in inDict:
        if type(inDict[key]) == np.ndarray:
            outDict['0'][key] = inDict[key][inDict['pol'] == False]
            outDict['1'][key] = inDict[key][inDict['pol'] == True]
        else:
            outDict['0'][key] = inDict[key]
            outDict['1'][key] = inDict[key]
    return outDict

'''
expecting startTime, stopTime or both
If given a single dataType dict, will just cut down all arrays by masking on the ts array. 
If given a larger container, will split down all that it finds, realigning timestamps.
If the container contains an info field, then the start and stopTime params
will be added.
'''
def cropTime(inDict, **kwargs):
    if isinstance(inDict, list):
        return [cropTime(inDictInst, **kwargs) for inDictInst in inDict]
    elif 'info' in inDict:
        outDict = {'info': inDict['info'].copy(),
                   'data': {}}
        for channelName in inDict['data'].keys():
            outDict['data'][channelName] = {}
            for dataTypeName in inDict['data'][channelName].keys():
                outDict['data'][channelName][dataTypeName] = cropTime(inDict['data'][channelName][dataTypeName], **kwargs)                
        rezeroTimestampsForImportedDicts(outDict)
        return outDict
    elif 'ts' in inDict:
        ts = inDict['ts']
        if not np.any(ts): # the dataset is empty
            return inDict
        startTime = kwargs.get('startTime', 
                    kwargs.get('minTime', 
                    kwargs.get('beginTime', 
                    kwargs.get('startTs', 
                    kwargs.get('minTs', 
                    kwargs.get('beginTs', 
                    ts[0]))))))
        stopTime = kwargs.get('stopTime', 
                   kwargs.get('maxTime', 
                   kwargs.get('endTime', 
                   kwargs.get('stopTs', 
                   kwargs.get('maxTs', 
                   kwargs.get('endTs', 
                   ts[-1]))))))
        if startTime == ts[0] and stopTime == ts[-1]:
            # No cropping to do - pass out the dict unmodified
            return inDict
        startIdx = np.searchsorted(ts, startTime, side='left')
        stopIdx = np.searchsorted(ts, stopTime, side='right')
        tsNew = ts[startIdx:stopIdx]
        if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
            tsNew = tsNew - startTime
        outDict = {'ts': tsNew}
        for fieldName in inDict.keys():
            if fieldName != 'ts':
                field = inDict[fieldName]
                try:
                    outDict[fieldName] = field[startIdx:stopIdx]
                except IndexError:
                    outDict[fieldName] = field.copy() # This might fail for certain data types
                except TypeError:
                    outDict[fieldName] = field # This might fail for certain data types
        if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
            tsOffsetOriginal = inDict.get('tsOffset', 0)
            outDict['tsOffset'] = tsOffsetOriginal - startTime
        return outDict
    else:
        # We assume that this is a datatype which doesn't contain ts, 
        # so we pass it out unmodified
        return inDict

# Could apply to any datatype with separate xy arrays; at the time of writing:
# dvs, flow
def cropSpaceXYArrays(inDict, **kwargs):
    if len(inDict['ts']) == 0: # no data to crop
        return inDict
    x = inDict['x']
    y = inDict['y']
    minX = kwargs.get('minX', np.min(x))
    maxX = kwargs.get('maxX', np.max(x))
    minY = kwargs.get('minY', np.min(y))
    maxY = kwargs.get('maxY', np.max(y))
    if (minX == np.min(x) and 
        maxX == np.max(x) and 
        minY == np.min(y) and 
        maxY == np.max(y)):
        # No cropping to do - pass out the dict unmodified
        return inDict
    selectedBool = np.logical_and(x >= minX, \
                                  np.logical_and(x <= maxX, \
                                          np.logical_and(y >= minY, y <= maxY)))
    return selectByBool(inDict, selectedBool)

def cropSpacePoint(inDict, **kwargs):
    if len(inDict['ts']) == 0: # no data to crop
        return inDict
    x = inDict['point'][:, 0]
    y = inDict['point'][:, 1]
    z = inDict['point'][:, 2]
    minX = kwargs.get('minX', np.min(x))
    maxX = kwargs.get('maxX', np.max(x))
    minY = kwargs.get('minY', np.min(y))
    maxY = kwargs.get('maxY', np.max(y))
    minZ = kwargs.get('minZ', np.min(z))
    maxZ = kwargs.get('maxZ', np.max(z))
    if (minX == np.min(x) and 
        maxX == np.max(x) and 
        minY == np.min(y) and 
        maxY == np.max(y) and 
        minZ == np.min(z) and 
        maxZ == np.max(z)):
        # No cropping to do - pass out the dict unmodified
        return inDict
    selectedBool = np.logical_and(x >= minX, \
                   np.logical_and(x <= maxX, \
                   np.logical_and(y >= minY, \
                   np.logical_and(y <= maxY, \
                   np.logical_and(z >= minZ, \
                                  z <= maxZ)))))
    return selectByBool(inDict, selectedBool)

def cropSpaceFrame(inDict, **kwargs):
    if len(inDict['ts']) == 0: # no data to crop
        return inDict
    firstFrame = inDict['frame'][0]
    minX = kwargs.get('minX', 0)
    maxX = kwargs.get('maxX', firstFrame.shape[1] - 1)
    minY = kwargs.get('minY', 0)
    maxY = kwargs.get('maxY', firstFrame.shape[0] - 1)
    if (minX == 0 and 
        maxX == firstFrame.shape[1] - 1 and 
        minY == 0 and 
        maxY == firstFrame.shape[0] - 1):
        # No cropping to do - pass out the dict unmodified
        return inDict
    framesCropped = [frame[minY:maxY+1, minX:maxX+1] for frame in inDict['frame']]
    return{**inDict, 'frame': framesCropped}

def cropSpace(inDict, **kwargs):
    if isinstance(inDict, list):
        return [cropSpace(inDictInst, **kwargs) for inDictInst in inDict]
    elif 'info' in inDict:
        outDict = {'info': inDict['info'].copy(),
                   'data': {}}
        for channelName in inDict['data'].keys():
            outDict['data'][channelName] = {}
            for dataTypeName in inDict['data'][channelName].keys():
                outDict['data'][channelName][dataTypeName] = cropSpace(inDict['data'][channelName][dataTypeName], **kwargs)
        # TODO: consider rezeroing space
        return outDict
    elif 'x' in inDict and 'y' in inDict:
        return cropSpaceXYArrays(inDict, **kwargs)
    elif 'frame' in inDict:
        return cropSpaceFrame(inDict, **kwargs)
    elif 'point' in inDict:
        return cropSpacePoint(inDict, **kwargs)
    else:
        # We assume that this is a datatype which doesn't contain x/y
        # so we pass it out unmodified
        # TODO: frame datatype could be cropped spatially but doesn't get caught by this method
        return inDict

def cropSpaceTime(inDict, **kwargs):
    return cropSpace(cropTime(inDict, **kwargs), **kwargs)

# synonyms
def cropSpatial(inDict, **kwargs):
    return cropSpace(inDict, **kwargs)

# synonyms
def cropTemporal(inDict, **kwargs):
    return cropTime(inDict, **kwargs)

# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def find_nearest(array, value):
    idx = np.searchsorted(array, value) # side="left" param is the default
    if idx > 0 and (idx == len(array) or
                    fabs(value - array[idx-1]) < fabs(value - array[idx])):
        return idx-1
    else:
        return idx

'''
getSamplesAtTimes takes a data-type dict and a parameter 'times',
and returns a dict with one sample for each element in times, 
being the sample in the original dict closest in time.
The samples in the output dict might be repeated.
'''
def getSamplesAtTimes(inDict, times, maxTimeDiff=None, allowDuplicates=False):
    ts = inDict['ts']
    ids = np.searchsorted(ts, times) # side="left" param is the default
    ids[ids == ts.shape[0]] = ts.shape[0] - 1 # if the index went beyond the final sample, then bring it backwards
    numSamples = ts.shape[0]
    shiftLeft = np.bitwise_and(
            ids > 0, \
            np.bitwise_or(
                ids == numSamples,
                np.absolute(times - ts[ids-1]) < np.absolute(times - ts[ids])))
    ids[shiftLeft] = ids[shiftLeft] - 1
    if maxTimeDiff is not None:
        timeDiffs = times - ts[ids]
        keepBool = timeDiffs <= maxTimeDiff 
        ids = ids[keepBool]
    outDict = {}
    if allowDuplicates == False:
        ids = np.unique(ids)
    for key in inDict.keys():
        if type(inDict[key]) == np.ndarray:
            if inDict[key].shape[0] == numSamples:
                outDict[key] = inDict[key][ids] 
            else:
                outDict[key] = inDict[key]
        elif type(inDict[key]) == list:
            if len(inDict[key]) == numSamples:
                outDict[key] = [inDict[key][idx] for idx in ids]
            else:
                outDict[key] = inDict[key]
        else:
            outDict[key] = inDict[key]
            
    return outDict

#------------------------------------------------------------------------------
'''
Merges multiple dicts then sorts by timestamp.
If strict = True (by default) then errors are raised if:
    a) Any of the list are None, or
    b) Any singleton keys have contradictory values an error is raised.
If not strict, then these problems are reported but the merge continues.
'''
def mergeDataTypeDicts(listOfDicts, **kwargs):
    outDict = {}
    strict = kwargs.get('strict', True)
    for inDict in listOfDicts:
        if inDict is None:
            problemMsg = 'List of dataType dicts contained "None"'
            if strict:
                raise ValueError(problemMsg)
            else:
                print(problemMsg)
                continue
        for key in inDict.keys():
            if key not in outDict:
                outDict[key] = inDict[key]
            else:
                if type(inDict[key]) == np.ndarray:
                    if inDict[key].shape[0] > 0:
                        outDict[key] = np.concatenate((outDict[key], inDict[key]))
                elif type(inDict[key]) == list:
                    if len(inDict[key]) > 0:
                        outDict[key] = outDict[key] + inDict[key]
                else: # assume singleton
                    if inDict[key] != outDict[key]:
                        problemMsg = 'Singleton keys found with contradictory values'
                        if strict:
                            raise ValueError(problemMsg)
                        else:
                            print(problemMsg)
    outDict = sortDataTypeDictByTime(outDict)
    return outDict
#------------------------------------------------------------------------------
def groupFlowTimeWindow(flowData, timeWindow):
    """
    Group the FLOW events based on a user-defined time window.

    Arguments:
        flowData {dict} -- dictionary of FLOW events as formatted by bimvee from event-driven library
        timeWindow {float} -- user-defined time window

    Returns:
        [list] -- list of dictionaries of the grouped FLOW events
    """
    flowDataGrouped = []
    timeWindowCount = int(round((max(flowData['ts']) - min(flowData['ts']))/timeWindow))
    for twc in range(0, timeWindowCount):
        flowDataCroppedTime = cropTime(flowData,
                                       startTime = flowData['ts'][0] + timeWindow * twc,
                                       stopTime = flowData['ts'][0] + timeWindow * (twc + 1),
                                       zeroTime = False)

        if len(flowDataCroppedTime['ts']) != 0:
            flowDataCroppedTime.update({'numSamples': len(flowDataCroppedTime['ts']), 'twc': twc})
            flowDataGrouped.append(flowDataCroppedTime)

    return flowDataGrouped

#------------------------------------------------------------------------------
def groupImuTimeWindow(imuData, timeWindow, flowData):
    """
    Group the IMU samples based on a user-defined time window, and taking also into account the FLOW events time windows.

    Arguments:
        imuData {dict} -- dictionary of IMU samples as formatted by bimvee from the event-driven library
        timeWindow {float} -- user-defined time window
        flowData {dict} -- dictionary of FLOW events as formatted by bimvee from the event-driven library

    Returns:
        [list] -- list of dictionaries of the grouped IMU samples
    """
    imuDataGrouped = []
    timeWindowCount = int(round((max(imuData['ts']) - min(imuData['ts']))/timeWindow))
    for twc in range(0, timeWindowCount):
        imuDataCroppedTime = cropTime(imuData,
                                      startTime = flowData['ts'][0] + timeWindow * twc,
                                      stopTime = flowData['ts'][0] + timeWindow * (twc + 1),
                                      zeroTime = False)

        if len(imuDataCroppedTime['ts']) != 0:
            imuDataCroppedTime.update({'numSamples': len(imuDataCroppedTime['ts']), 'twc': twc})
            imuDataGrouped.append(imuDataCroppedTime)

    return imuDataGrouped



# -*- coding: utf-8 -*-

"""
Copyright (C) 2021 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

The importTopic function receives a list of messages and returns
a dict with one field for each data field in the message, where the field
will contain an appropriate iterable to contain the interpretted contents of each message.
In some cases, static info is repeated in each message; in which case a field may not contain an iterable. 

This function imports the ros message type defined at:
http://docs.ros.org/api/sensor_msgs/html/msg/Imu.html
"""

#%%

from tqdm import tqdm
import numpy as np

from .common import unpackRosString, unpackRosTimestamp, unpackRosFloat64Array

def importTopic(msgs, **kwargs):
    '''
    ros message is defined here:
        http://docs.ros.org/en/api/sensor_msgs/html/msg/Illuminance.html
    the result is are np arrays of float64 for:
        illuminance (1 col)
    '''
    sizeOfArray = 1024
    tsAll = np.zeros((sizeOfArray), dtype=np.float64)
    illuminanceAll = np.zeros((sizeOfArray, 1), dtype=np.float64)
    for idx, msg in enumerate(tqdm(msgs, position=0, leave=True)):
        if sizeOfArray <= idx:
            tsAll = np.append(tsAll, np.zeros((sizeOfArray), dtype=np.float64))
            illuminanceAll = np.concatenate((illuminanceAll, np.zeros((sizeOfArray, 1), dtype=np.float64)))
            sizeOfArray *= 2
        data = msg['data']
        #seq = unpack('=L', data[0:4])[0]
        tsAll[idx], ptr = unpackRosTimestamp(data, 4)
        frame_id, ptr = unpackRosString(data, ptr)
        illuminanceAll[idx, :], ptr = unpackRosFloat64Array(data, 1, ptr)
        ptr += 8 # Skip the covariance matrix
    numEvents = idx + 1
    # Crop arrays to number of events
    tsAll = tsAll[:numEvents]
    illuminanceAll = illuminanceAll[:numEvents]
    outDict = {
        'ts': tsAll,
        'illuminance': illuminanceAll,
        }
    return outDict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
An example of 'aer2' format in the wild is:
    
https://github.com/VLOGroup/dvs-reconstruction/blob/master/data/sample_data.tar.bz2

It's a text file, no headers, where each event encodes one line, in the  format:
    
ts x y pol

importAer2 opens such a file (pass with filePathOrName parameter) 
and returns a dict in this format:
{'info': {},
 'data': {
         ch0: {
               dvs: {
                     'ts': np.array of float in seconds 
                     'x': np.array of np.uint16 in pixels
                     'y': np.array of np.uint16 in pixels
                     'pol': np.array of bool -- 1 = ON event 
                    }}}}
"""

import numpy as np
from tqdm import tqdm 

# Local imports
from .timestamps import zeroTimestampsForADataType

def inferDim(array):
    dimStops = np.array([32, 64, 128, 180, 240, 256, 260, 304, 320, 346, 480, 640, 720, 1080, 1280, 1920], dtype=np.uint16)
    idx = np.searchsorted(dimStops, np.max(array))
    try:
        return dimStops[idx]
    except IndexError:
        return np.max(array) + 1
    
def importAer2(**kwargs):
    filePathOrName = kwargs['filePathOrName']
    print('Attempting to import ' + filePathOrName + ' as aer2')
    sizeOfArray = 1024
    ts = np.zeros((sizeOfArray), dtype=float)
    x = np.zeros((sizeOfArray), dtype=np.uint16)
    y = np.zeros((sizeOfArray), dtype=np.uint16)
    pol = np.zeros((sizeOfArray), dtype=bool)

    with open(filePathOrName, 'r') as file:
        for idx, line in enumerate(tqdm(file)):
            if idx == sizeOfArray:
                ts = np.concatenate((ts, np.zeros((sizeOfArray), dtype=float)))
                x = np.concatenate((x, np.zeros((sizeOfArray), dtype=np.uint16)))
                y = np.concatenate((y, np.zeros((sizeOfArray), dtype=np.uint16)))
                pol = np.concatenate((pol, np.zeros((sizeOfArray), dtype=bool)))
                sizeOfArray *= 2
            lineSplit = line.split()
            ts[idx] = float(lineSplit[0])
            x[idx] = int(lineSplit[1])
            y[idx] = int(lineSplit[2])
            pol[idx] = int(lineSplit[3])
        numEvents = idx + 1
    dvsDict = {'ts': ts[:numEvents] / 1000000,
           'x': x[:numEvents],
           'y': y[:numEvents],
           'pol': pol[:numEvents],
           'dimX': inferDim(x),
           'dimY': inferDim(y)
           }

    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)): 
        zeroTimestampsForADataType(dvsDict)
    outDict = {
    'info': {'filePathOrName':filePathOrName,
        'fileFormat': 'aer2'},
    'data': {
        'ch0': {
            'dvs': dvsDict
            }
        }
    }
    print('Done.')
    return outDict
